from pathlib import Path
from typing import cast

from pytest import MonkeyPatch
from pytest_mock import MockerFixture

from sunset import AutoLoader, Key, Settings
from sunset.autoloader import MonitorForChange, MonitorProtocol
from sunset.timer import TimerProtocol

from .test_timer import MockTimer


class DummyMonitor(MonitorForChange):
    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass


class ExampleSettings(Settings):
    key_str = Key(default="")


class TestMonitor:
    def test_monitor_triggers_callback_on_change(
        self, mocker: MockerFixture, tmp_path: Path
    ) -> None:
        callback = mocker.stub()
        path = mocker.patch("sunset.autoloader.Path", mocker.Mock(Path))
        mstats = mocker.Mock()
        mstats.st_mtime = 0
        path.stat.return_value = mstats
        mocker.patch("sunset.autoloader.PersistentTimer", MockTimer)
        monitor = MonitorForChange(path, callback)
        monitor.start()
        timer = monitor._timer
        assert timer is not None
        timer = cast(MockTimer, timer)

        # Simulate timer ticks by advancing time
        timer.advanceTime(1.0)
        path.stat.assert_called()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        mstats.st_mtime = 1

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_called_once()
        callback.reset_mock()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        mstats.st_mtime = 2

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_called_once()
        callback.reset_mock()

        monitor.stop()

    def test_monitor_supports_live_file_creation(
        self, mocker: MockerFixture, tmp_path: Path
    ) -> None:
        path = mocker.patch("sunset.autoloader.Path", mocker.Mock(Path))
        path.stat = mocker.stub()
        path.stat.side_effect = OSError("File not found")

        mocker.patch("sunset.autoloader.PersistentTimer", MockTimer)

        callback = mocker.stub()
        monitor = MonitorForChange(path, callback)
        monitor.start()

        timer = monitor._timer
        assert timer is not None
        timer = cast(MockTimer, timer)

        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        # Simulate file creation.
        path.stat.side_effect = None
        mstats = mocker.Mock()
        mstats.st_mtime = 0
        path.stat.return_value = mstats

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_not_called()

        timer.advanceTime(1.0)
        path.stat.assert_called_once()
        path.stat.reset_mock()
        callback.assert_called_once()
        monitor.stop()

    def test_starting_and_stopping_a_monitor_is_idempotent(
        self, mocker: MockerFixture, tmp_path: Path
    ) -> None:
        path = tmp_path / "test.conf"
        callback = mocker.Mock()
        mock_timer = mocker.Mock(TimerProtocol)

        mocker.patch("sunset.autoloader.PersistentTimer", mock_timer)
        monitor = MonitorForChange(path, callback)

        mock_timer().start.assert_not_called()

        monitor.start()
        mock_timer().start.assert_called_once()
        mock_timer().start.reset_mock()

        monitor.start()
        mock_timer().start.assert_not_called()
        mock_timer().cancel.assert_not_called()

        monitor.stop()
        mock_timer().cancel.assert_called_once()
        mock_timer().cancel.reset_mock()

        monitor.stop()
        mock_timer().cancel.assert_not_called()


class TestAutoLoader:
    def test_autoloader_loads_settings_on_init(self, tmp_path: Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings_file.write_text("[main]\nkey_str = test\n")

        settings = ExampleSettings()
        assert settings.key_str.get() == ""

        with AutoLoader(settings, settings_file, _monitor_factory=DummyMonitor):
            assert settings.key_str.get() == "test"

    def test_autoloader_reloads_settings_on_change(self, tmp_path: Path) -> None:
        settings_file = tmp_path / "test.conf"

        settings = ExampleSettings()

        with AutoLoader(
            settings, settings_file, _monitor_factory=DummyMonitor
        ) as autoloader:
            assert settings.key_str.get() == ""

            # Simulate a change in the settings file.
            settings_file.write_text("[main]\nkey_str = new value\n")
            autoloader._monitor.triggerCallback()

            assert settings.key_str.get() == "new value"

    def test_autoloader_start_and_stops_monitor(
        self, mocker: MockerFixture, tmp_path: Path
    ) -> None:
        monitor = mocker.Mock(spec=MonitorProtocol)
        monitor().start = mocker.stub()
        monitor().stop = mocker.stub()
        monitor().trigger_callback = mocker.stub()

        settings_file = tmp_path / "test.conf"
        settings_file.write_text("")

        autoloader = AutoLoader(Settings(), settings_file, _monitor_factory=monitor)

        # The file monitor is started on init.
        monitor().start.assert_called_once()
        monitor().start.reset_mock()
        monitor().stop.assert_not_called()

        with autoloader:
            # The file monitor is started on entering the context manager.
            monitor().start.assert_called_once()
            monitor().start.reset_mock()
            monitor().stop.assert_not_called()

        # The file monitor is stopped on exiting the context manager.
        monitor().stop.assert_called_once()
        monitor().stop.reset_mock()
        monitor().start.assert_not_called()

        # This remains true if the context manager is entered again later on.
        with autoloader:
            monitor().start.assert_called_once()
            monitor().start.reset_mock()
            monitor().stop.assert_not_called()

        monitor().stop.assert_called_once()
        monitor().stop.reset_mock()
        monitor().start.assert_not_called()

        # The monitor is stopped when the autoloader is garbage collected.
        monitor.reset_mock()  # Clear the mock's references to the autoloader object.
        del autoloader
        monitor().start.assert_not_called()
        monitor().stop.assert_called_once()

    def test_expand_user(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        def expanduser(path: Path) -> Path:
            if not str(path).startswith("~"):
                return path
            return Path("HOME") / str(path).lstrip("~").lstrip("/")

        monkeypatch.setattr(Path, "expanduser", expanduser)

        autoloader1 = AutoLoader(
            ExampleSettings(), "/no/tilde", _monitor_factory=DummyMonitor
        )
        assert str(autoloader1.path()) == "/no/tilde"

        autoloader2 = AutoLoader(
            ExampleSettings(), "~/with/tilde", _monitor_factory=DummyMonitor
        )
        assert str(autoloader2.path()) == "HOME/with/tilde"



import threading
import weakref
from collections.abc import Callable, Iterator, MutableMapping, MutableSet
from types import MethodType
from typing import Any, TypeVar, cast

_T = TypeVar("_T")
_C = TypeVar("_C", bound=Callable[..., Any])


class NonHashableSet(MutableSet[_T]):
    """
    An implementation of a set that can contain non-hashable elements.

    Elements that are not hashable are distinguished by their id.
    """

    _contents: MutableMapping[int, _T]
    _lock: threading.RLock

    def __init__(self, mapping_type: type[MutableMapping[int, _T]] = dict) -> None:
        super().__init__()

        self._contents = mapping_type()
        self._lock = threading.RLock()

    @staticmethod
    def _computeHash(value: object) -> int:
        # Try to return the normal hash for the value if possible. This is
        # because hash is more selective than id. For instance! If value is a
        # weakref, then hash properly identifies two distinct weakrefs to the
        # same object as equivalent, whereas id does not.
        # Sadly testing whether the value implements the Hashable proto does not
        # suffice. Dataclasses, for instance, implement the Hashable proto but
        # are not necessarily hashable.

        try:
            return hash(value)
        except TypeError:
            return id(value)

    def add(self, value: _T) -> None:
        with self._lock:
            self._contents[self._computeHash(value)] = value

    def discard(self, value: _T) -> None:
        try:
            with self._lock:
                del self._contents[self._computeHash(value)]
        except KeyError:
            pass

    def __contains__(self, obj: object) -> bool:
        return self._computeHash(obj) in self._contents

    def __iter__(self) -> Iterator[_T]:
        with self._lock:
            yield from self._contents.values()

    def __len__(self) -> int:
        return len(self._contents)


class WeakNonHashableSet(NonHashableSet[_T]):
    """
    An implementation of a weak set that can contain non-hashable elements.

    Elements that are not hashable are distinguished by their id.
    """

    def __init__(self) -> None:
        super().__init__(mapping_type=weakref.WeakValueDictionary)


class WeakCallableSet(MutableSet[_C]):
    """
    A set-like type that stores callables without keeping references to them.

    The callables must have the same argument types and return type.
    """

    _content: NonHashableSet[weakref.ReferenceType[_C]]

    def __init__(self) -> None:
        super().__init__()

        # We can't just use a WeakNonHashableSet because it does not know how to
        # take references to methods. Instead we use a regular NonHashableSet
        # and manually take the proper type of weak reference when adding a
        # callable to the set.

        self._content = NonHashableSet()

    def add(self, value: _C) -> None:
        if isinstance(value, MethodType):
            r: weakref.ReferenceType[_C] = weakref.WeakMethod(
                cast(_C, value), self._onExpire
            )

        else:
            r = weakref.ref(value, self._onExpire)

        self._content.add(r)

    def __contains__(self, value: object) -> bool:
        return any(self._isSameCallable(candidate, value) for candidate in self)

    def __iter__(self) -> Iterator[_C]:
        for ref in self._content:
            value = ref()
            if value is not None:
                yield value

    def __len__(self) -> int:
        return len(self._content)

    def discard(self, value: _C) -> None:
        to_discard: list[weakref.ReferenceType[_C]] = []
        for ref in self._content:
            callable_ = ref()
            if callable_ is not None and self._isSameCallable(callable_, value):
                to_discard.append(ref)

        for ref in to_discard:
            self._content.discard(ref)

    @staticmethod
    def _isSameCallable(callable1: _C, callable2: object) -> bool:
        if isinstance(callable1, MethodType) and isinstance(callable2, MethodType):
            return (
                callable1.__self__ is callable2.__self__
                and callable1.__name__ == callable2.__name__
            )

        return callable1 is callable2

    def _onExpire(self, ref: weakref.ReferenceType[_C]) -> None:
        self._content.discard(ref)



import dataclasses
import inspect
import sys
import weakref
from collections.abc import Callable, Iterable, MutableSet
from typing import TYPE_CHECKING, Any, cast

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:
    from typing import Self

if TYPE_CHECKING:  # pragma: no cover
    from types import GenericAlias

from sunset.lock import SettingsLock
from sunset.notifier import Notifier
from sunset.protocols import BaseField, Field, ItemTemplate, UpdateNotifier
from sunset.sets import WeakNonHashableSet
from sunset.stringutils import collate_by_prefix, split_on


class Bunch(BaseField):
    """
    A collection of related Keys.

    Under the hood, a Bunch is a dataclass, and can be used in the same manner,
    i.e. by defining attributes directly on the class itself.

    Example:

    >>> from sunset import Bunch, Key
    >>> class Appearance(Bunch):
    ...     class Font(Bunch):
    ...         name: Key[str] = Key(default="Arial")
    ...         size: Key[int] = Key(default=14)
    ...     main_font: Font      = Font()
    ...     secondary_font: Font = Font()
    >>> appearance = Appearance()
    >>> appearance.main_font.name.get()
    'Arial'
    >>> appearance.secondary_font.name.get()
    'Arial'
    >>> appearance.main_font.name.set("Times New Roman")
    True
    >>> appearance.secondary_font.name.set("Calibri")
    True
    >>> appearance.main_font.name.get()
    'Times New Roman'
    >>> appearance.secondary_font.name.get()
    'Calibri'
    """

    _parent_ref: weakref.ref["Bunch"] | None
    _children_set: MutableSet["Bunch"]
    _fields: dict[str, Field]
    _update_notifier: Notifier[[UpdateNotifier]]

    def __new__(cls) -> Self:
        # Build and return a dataclass constructed from this class. Keep a
        # reference to that dataclass as a private class attribute, so that we
        # only construct it once. This allows type identity checks (as in
        # "type(a) is type(b)") to work.

        dataclass_attr = "__DATACLASS_CLASS"
        orig_class_attr = "__ORIG_CLASS"

        dataclass_class: type[Self] | None = None
        if dataclasses.is_dataclass(cls):
            # If this class is already a dataclass, then no need to construct a new one.
            # Just use this one directly.
            dataclass_class = cls
        else:
            # Else use the dataclass recorded on this class, if there is one. Note the
            # use of vars(), in order to look up the dataclass for this specific
            # class, and not one of its parents.
            dataclass_class = vars(cls).get(dataclass_attr, None)

        if dataclass_class is None:
            # We haven't yet constructed a dataclass from this class. Construct
            # one here.

            cls_parents = cls.__bases__
            dataclass_fields: list[tuple[str, type | GenericAlias, ItemTemplate]] = []
            potential_fields = [
                (name, getattr(cls, name, None))
                for name in dir(cls)
                if not name.startswith("__")
            ]

            for name, attr in potential_fields:
                if inspect.isclass(attr):
                    if attr.__name__ == name:
                        # This is probably a class definition that just happens
                        # to be located inside the containing Bunch definition.
                        # This is fine.

                        continue

                    msg = (
                        f"Field '{name}' in the definition of '{cls.__name__}' is"
                        " uninstantiated. Did you forget the parentheses?"
                    )
                    raise TypeError(msg)

                if isinstance(attr, ItemTemplate):
                    # Safety check: make sure the user isn't accidentally overriding an
                    # existing attribute. We do however allow overriding an attribute
                    # with an attribute of the same type, which allows the user to
                    # override a Key with a Key of the same type but a different
                    # default value, for instance.

                    for cls_parent in cls_parents:
                        if (
                            parent_attr := getattr(cls_parent, name, None)
                        ) is not None and type(parent_attr) is not type(attr):
                            msg = (
                                f"Field '{name}' in the definition of"
                                f" '{cls.__name__}' overrides attribute of the"
                                " same name declared in parent class"
                                f" '{cls_parent.__name__}'; consider renaming"
                                f" this field to '{name}_' for instance"
                            )
                            raise TypeError(msg)

                    # Create a proper field from the attribute.

                    field = dataclasses.field(default_factory=attr._newInstance)  # noqa: SLF001
                    dataclass_fields.append((name, attr._typeHint(), field))  # noqa: SLF001

            # Create a dataclass based on this class. Note that we will be
            # providing our own __init__() override below.

            kwargs: dict[str, Any] = (
                {} if sys.version_info < (3, 12) else {"module": cls.__module__}
            )
            dataclass_class = cast(
                type[Self],
                dataclasses.make_dataclass(
                    cls.__qualname__,
                    dataclass_fields,
                    init=False,
                    bases=(cls, *cls_parents),
                    **kwargs,
                ),
            )

            # And store it on the class itself so it can be reused if this class is
            # instantiated again.

            setattr(cls, dataclass_attr, dataclass_class)

            # Also keep a reference to the original class, so that's not lost.

            setattr(dataclass_class, orig_class_attr, cls)

        # Create an instance of the dataclass.

        new_cls = super().__new__(dataclass_class)

        # Set up the fields that were identified above as instance attributes.

        new_cls.__setup_fields__()
        return new_cls

    def __setup_fields__(self) -> None:
        # Set up the Bunch fields as instance attributes.
        # First, look up internal dataclass attribute names.

        fields_attr: str = getattr(dataclasses, "_FIELDS")  # noqa: B009
        field_type: Any = getattr(dataclasses, "_FIELD")  # noqa: B009
        fields: dict[str, dataclasses.Field[Any]] = getattr(self, fields_attr, {})

        # Then look up the fields stored in the dataclass.

        for field in fields.values():
            if (
                getattr(field, "_field_type", None) is not field_type
            ):  # pragma: no cover
                continue

            if field.default_factory is dataclasses.MISSING:  # pragma: no cover
                continue

            setattr(self, field.name, field.default_factory())

    def __init__(self) -> None:
        super().__init__()

        self._parent_ref = None
        self._children_set = WeakNonHashableSet[Bunch]()
        self._fields = {}
        self._update_notifier = Notifier()

        for label, field in vars(self).items():
            if isinstance(field, Field):
                self._fields[label] = field
                field.meta().update(label=label, container=self)
                field._update_notifier.add(self._update_notifier.trigger)  # noqa: SLF001

        self.__post_init__()

    def __post_init__(self) -> None:
        """
        DEPRECATED. Will be removed in v1.0.

        Use __init__() instead.
        """

    @SettingsLock.with_write_lock
    def setParent(self, parent: Self | None) -> None:
        """
        Makes the given Bunch the parent of this one. If None, remove this
        Bunch's parent, if any.

        All the Key, List and Bunch fields defined on this Bunch instance
        will be recursively reparented to the corresponding Key / List / Bunch
        field on the given parent.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Bunch that will become this Bunch's parent, or
                None. The parent Bunch must have the same type as this
                Bunch.

        Note:
            A Bunch and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None and type(self) is not type(parent):  # pragma: no cover
            return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children_set.discard(self)  # noqa: SLF001

        if parent is None:
            self._parent_ref = None
        else:
            self._parent_ref = weakref.ref(parent)
            parent._children_set.add(self)  # noqa: SLF001

        for label, field in self._fields.items():
            if parent is None:
                field.setParent(None)
                continue

            parent_field = parent._fields.get(label)
            if parent_field is None:  # pragma: no cover
                # This is a safety check, but it shouldn't happen. By
                # construction self should be of the same type as parent, so
                # they should have the same attributes.

                continue

            assert type(field) is type(parent_field)  # noqa: S101
            field.setParent(parent_field)

    @SettingsLock.with_read_lock
    def parent(self) -> Self | None:
        """
        Returns the parent of this Bunch, if any.

        Returns:
            A Bunch instance of the same type as this one, or None.
        """

        # Make the type of self._parent_ref more specific for the purpose of
        # type checking.

        parent = cast(weakref.ref[Self] | None, self._parent_ref)
        return parent() if parent is not None else None

    @SettingsLock.with_read_lock
    def children(self) -> Iterable[Self]:
        """
        Returns an iterable with the Bunch instances that have this Bunch as their
        parent.

        Returns:
            An iterable of Bunch instances of the same type as this one.
        """

        return [cast(Self, child) for child in self._children_set]

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this Bunch is updated. A Bunch
        is considered updated when any of its fields is updated.

        The callback will be called with as its argument whichever field was
        just updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notifier.add(callback)

    @SettingsLock.with_read_lock
    def isSet(self) -> bool:
        """
        Indicates whether this Bunch holds any field that is set.

        Returns:
            True if any field set on this Bunch is set, else False.
        """

        return any(field.isSet() for field in self._fields.values())

    @SettingsLock.with_write_lock
    def clear(self) -> None:
        [field.clear() for field in self._fields.values()]

    @SettingsLock.with_read_lock
    def dumpFields(self) -> Iterable[tuple[str, str | None]]:
        """
        Internal.
        """

        ret: list[tuple[str, str | None]] = []
        sep = self._PATH_SEPARATOR
        if not self.skipOnSave():
            for label, field in sorted(self._fields.items()):
                ret.extend(
                    (label + sep + path if path else label, item)
                    for path, item in field.dumpFields()
                )

        return ret

    @SettingsLock.with_write_lock
    def restoreFields(self, fields: Iterable[tuple[str, str | None]]) -> bool:
        """
        Internal.
        """

        any_change = False
        by_field = collate_by_prefix(fields, split_on(self._PATH_SEPARATOR))

        with self._update_notifier.inhibit():
            for field_name, field in self._fields.items():
                if field_name not in by_field:
                    # If the field is not in the fields to restore, clear it.

                    any_change = any_change or field.isSet()
                    field.clear()

                else:
                    any_change = field.restoreFields(by_field[field_name]) or any_change

        return any_change

    def _typeHint(self) -> type:
        return type(self)

    def _newInstance(self) -> Self:
        """
        Internal. Returns a new instance of this Bunch with the same fields.

        Returns:
            A Bunch instance.
        """

        return self.__class__()



import sys
import weakref
from collections.abc import Callable, Iterable, Iterator, MutableSequence
from enum import Enum, auto
from types import GenericAlias
from typing import Any, SupportsIndex, TypeVar, cast, overload

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:
    from typing import Self

from sunset.bunch import Bunch
from sunset.key import Key
from sunset.lock import SettingsLock
from sunset.notifier import Notifier
from sunset.protocols import BaseField, UpdateNotifier
from sunset.sets import WeakNonHashableSet
from sunset.stringutils import collate_by_prefix, split_on

ListItemT = TypeVar("ListItemT", bound=Bunch | Key[Any])


class IterOrder(Enum):
    NO_PARENT = auto()
    PARENT_FIRST = auto()
    PARENT_LAST = auto()


class List(MutableSequence[ListItemT], BaseField):
    """
    A list-like container for Keys or Bunches of a given type, to be used in a
    Settings' definition.

    It is type-compatible with standard Python lists and supports indexing,
    insertion, appending, etc.

    In addition, it offers support for update notification callbacks, and
    inheritance. The inheritance is used when iterating on the List's contents
    using the :meth:`iter()` method, which iterates on a List and, optionally,
    its parents.

    Instead of creating a new instance of the contained type and inserting or
    appending it, you can use the :meth:`appendOne()` or :meth:`insertOne()`
    methods to do it in one step.

    Args:
        template: A Key or a Bunch *instance* that represents the items that
            will be contained in this List. (The template itself will not be
            added to the List.)

        order: One of `List.NO_PARENT`, `List.PARENT_FIRST` or
            `List.PARENT_LAST`. Sets the default iteration order used in
            :meth:`iter()` when not otherwise specified. Default:
            `List.NO_PARENT`.

    Example:

    >>> from sunset import Bunch, Key, List, Settings
    >>> class ExampleSettings(Settings):
    ...     class ExampleBunch(Bunch):
    ...         a: Key[str] = Key(default="")
    ...     key_list: List[Key[int]]         = List(Key(default=0))
    ...     bunch_list: List[ExampleBunch] = List(ExampleBunch())
    >>> settings = ExampleSettings()
    >>> settings.bunch_list
    []
    >>> settings.bunch_list.appendOne().a.set("demo")
    True
    >>> settings.bunch_list
    [ExampleSettings.ExampleBunch(a=<Key[str]:demo>)]
    >>> settings.key_list
    []
    >>> settings.key_list.appendOne().set(12)
    True
    >>> settings.key_list
    [<Key[int]:12>]
    """

    NO_PARENT = IterOrder.NO_PARENT
    PARENT_FIRST = IterOrder.PARENT_FIRST
    PARENT_LAST = IterOrder.PARENT_LAST

    _contents: list[ListItemT]
    _parent_ref: weakref.ref["List[ListItemT]"] | None
    _children_ref: WeakNonHashableSet["List[ListItemT]"]
    _iter_order: IterOrder
    _update_notifier: Notifier[[UpdateNotifier]]
    _template: ListItemT

    def __init__(
        self, template: ListItemT, order: IterOrder = IterOrder.NO_PARENT
    ) -> None:
        super().__init__()

        self._contents = []

        self._parent_ref = None
        self._children_ref = WeakNonHashableSet()
        self._iter_order = order
        self._update_notifier = Notifier()
        self._template = template

    @SettingsLock.with_write_lock
    def insert(self, index: SupportsIndex, value: ListItemT) -> None:
        self._contents.insert(index, value)
        self._relabelItems()
        self._update_notifier.trigger(self)

    @overload
    def __getitem__(self, index: SupportsIndex) -> ListItemT: ...

    @overload
    def __getitem__(self, index: slice) -> list[ListItemT]: ...

    @SettingsLock.with_read_lock
    def __getitem__(self, index: SupportsIndex | slice) -> ListItemT | list[ListItemT]:
        return self._contents[index]

    @overload
    def __setitem__(self, index: SupportsIndex, value: ListItemT) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[ListItemT]) -> None: ...

    @SettingsLock.with_write_lock
    def __setitem__(
        self, index: SupportsIndex | slice, value: ListItemT | Iterable[ListItemT]
    ) -> None:
        self._clearMetadata(self._contents[index])
        if isinstance(index, slice):
            assert isinstance(value, Iterable)  # noqa: S101
            self._contents[index] = value

        else:
            assert isinstance(index, SupportsIndex)  # noqa: S101
            assert not isinstance(value, Iterable)  # noqa: S101
            self._contents[index] = value

        self._relabelItems()
        self._update_notifier.trigger(self)

    @SettingsLock.with_write_lock
    def __delitem__(self, index: SupportsIndex | slice) -> None:
        self._clearMetadata(self._contents[index])
        del self._contents[index]
        self._relabelItems()
        self._update_notifier.trigger(self)

    @SettingsLock.with_write_lock
    def extend(self, values: Iterable[ListItemT]) -> None:
        self._contents.extend(values)
        self._relabelItems()
        self._update_notifier.trigger(self)

    @SettingsLock.with_write_lock
    def append(self, value: ListItemT) -> None:
        self.extend((value,))

    @SettingsLock.with_write_lock
    def __iadd__(self, values: Iterable[ListItemT]) -> Self:
        self.extend(values)
        return self

    @SettingsLock.with_write_lock
    def clear(self) -> None:
        del self[:]

    def _clearMetadata(self, fields: ListItemT | list[ListItemT]) -> None:
        for field in fields if isinstance(fields, list) else [fields]:
            field.meta().clear()
            field._update_notifier.discard(self._update_notifier.trigger)  # noqa: SLF001

    def __len__(self) -> int:
        return len(self._contents)

    def _newItem(self) -> ListItemT:
        return self._template._newInstance()  # noqa: SLF001

    @SettingsLock.with_read_lock
    def isSet(self) -> bool:
        """
        Indicates whether this List holds any item that is set.

        Returns:
            True if any item contained in this List is set, else False.
        """

        return any(item.isSet() for item in self._contents)

    @SettingsLock.with_write_lock
    def appendOne(self) -> ListItemT:
        """
        Creates a new item of the type contained in this List, appends it to
        this List, and returns it.

        Returns:
            An instance of the item type contained in this List.
        """

        item = self._newItem()
        self.append(item)
        return item

    @SettingsLock.with_write_lock
    def insertOne(self, index: int) -> ListItemT:
        """
        Creates a new item of the type contained in this List, inserts it in
        this List at the given index, and returns it.

        Args:
            index: The list index where to insert the new item.

        Returns:
            An instance of the item type contained in this List.
        """

        item = self._newItem()
        self.insert(index, item)
        return item

    def _relabelItems(self) -> None:
        for i, item in enumerate(self._contents):
            item.meta().update(label=self._labelForIndex(i), container=self)
            item._update_notifier.add(self._update_notifier.trigger)  # noqa: SLF001

    @staticmethod
    def _labelForIndex(index: SupportsIndex) -> str:
        return str(int(index) + 1)

    @staticmethod
    def _indexForLabel(label: str) -> int | None:
        if not label.isdigit():
            return None
        index = int(label) - 1
        return index if index >= 0 else None

    @SettingsLock.with_write_lock
    def setParent(self, parent: Self | None) -> None:
        """
        Makes the given List the parent of this one. If None, remove this List's
        parent, if any.

        Having a parent does not affect a List's behavior outside of the
        :meth:`iter()` method.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a List that will become this List's parent, or
                None. The parent List must have the same type as this List.

        Note:
            A List and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None and type(self) is not type(parent):  # pragma: no cover
            return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children_ref.discard(self)  # noqa: SLF001

        if parent is None:
            self._parent_ref = None
        else:
            self._parent_ref = weakref.ref(parent)
            parent._children_ref.add(self)  # noqa: SLF001

    @SettingsLock.with_read_lock
    def iter(self, order: IterOrder | None = None) -> Iterable[ListItemT]:
        """
        Returns the elements contained in this List, and optionally in its parents, if
        any.

        Args:
            order: One of `List.NO_PARENT`, `List.PARENT_FIRST`,
                `List.PARENT_LAST`, or None.

                - If `List.NO_PARENT`, this method only returns the contents of this
                  List instance.

                - If `List.PARENT_FIRST`, it recursively returns items from this List's
                  parents, if any, then from this List itself.

                - If `List.PARENT_LAST`, it returns items from this List itself, then
                  recursively from its parents, if any.

                - If None, the order set on the List itself at creation time
                  will be used.

                Default: None.

        Returns:
            An iterable on the items contained in this List and optionally its parents.

        Example:

        >>> from sunset import Key, List
        >>> show = lambda l: [key.get() for key in l]
        >>> parent = List(Key(default=0))
        >>> parent.appendOne().set(1)
        True
        >>> parent.appendOne().set(2)
        True
        >>> child = List(Key(default=0))
        >>> child.appendOne().set(3)
        True
        >>> child.appendOne().set(4)
        True
        >>> show(parent)
        [1, 2]
        >>> show(child)
        [3, 4]
        >>> child.setParent(parent)
        >>> show(child.iter())
        [3, 4]
        >>> show(child.iter(order=List.PARENT_FIRST))
        [1, 2, 3, 4]
        >>> show(child.iter(order=List.PARENT_LAST))
        [3, 4, 1, 2]
        """

        ret: list[ListItemT] = []
        parent = self.parent()

        if order is None:
            order = self._iter_order

        if parent is not None and order == IterOrder.PARENT_FIRST:
            ret.extend(parent.iter(order))

        ret.extend(self._contents)

        if parent is not None and order == IterOrder.PARENT_LAST:
            ret.extend(parent.iter(order))

        return ret

    def __iter__(self) -> Iterator[ListItemT]:
        items = self.iter(IterOrder.NO_PARENT)
        yield from items

    @SettingsLock.with_read_lock
    def parent(self) -> Self | None:
        """
        Returns the parent of this List, if any.

        Returns:
            A List instance of the same type as this one, or None.
        """

        # Make the type of self._parent_ref more specific for the purpose of
        # type checking.

        parent = cast(weakref.ref[Self] | None, self._parent_ref)
        return parent() if parent is not None else None

    @SettingsLock.with_read_lock
    def children(self) -> Iterable[Self]:
        """
        Returns an iterable with the List instances that have this List as their parent.

        Returns:
            An iterable of List instances of the same type as this one.
        """

        return [cast(Self, child) for child in self._children_ref]

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this List, *or* any item contained
        in this List, is updated.

        The callback will be called with whichever entity was updated as its
        argument: this List, or one of its items or sub-items.

        Adding new items to or deleting items from a List is considered an
        update of that List, not of the elements in question.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notifier.add(callback)

    @SettingsLock.with_read_lock
    def dumpFields(self) -> Iterable[tuple[str, str | None]]:
        """
        Internal.
        """

        ret: list[tuple[str, str | None]] = []
        sep = self._PATH_SEPARATOR
        if not self.skipOnSave():
            for i, item in enumerate(self._contents, start=1):
                label = str(i)
                if not item.isSet():
                    ret.append((label, None))
                else:
                    ret.extend(
                        (label + sep + path if path else label, child_item)
                        for path, child_item in item.dumpFields()
                    )

        return ret

    @SettingsLock.with_write_lock
    def restoreFields(self, fields: Iterable[tuple[str, str | None]]) -> bool:
        """
        Internal.
        """

        any_change = False

        by_label = collate_by_prefix(fields, split_on(self._PATH_SEPARATOR))
        by_index = {
            index: value
            for label, value in by_label.items()
            if (index := self._indexForLabel(label)) is not None
        }
        max_index = max(by_index.keys(), default=-1)

        with self._update_notifier.inhibit():
            if self._lastIndex > max_index:
                del self[max_index + 1 :]
                any_change = True
            else:
                any_change = self._ensureIndexExists(max_index)

            for i, field in enumerate(self):
                any_change = field.restoreFields(by_index.get(i, [])) or any_change

        return any_change

    @SettingsLock.with_write_lock
    def _ensureIndexExists(self, index: int) -> bool:
        missing_count = index - self._lastIndex

        if missing_count > 0:
            self.extend(self._newItem() for _ in range(missing_count))
            return True

        return False

    @property
    def _lastIndex(self) -> int:
        return len(self) - 1

    def _typeHint(self) -> GenericAlias:
        return GenericAlias(type(self), type(self._template))

    def _newInstance(self) -> Self:
        """
        Internal. Returns a new instance of this List capable of holding the
        same type.

        Returns:
            A new List.
        """

        return self.__class__(template=self._template, order=self._iter_order)

    def __repr__(self) -> str:
        parent = self.parent()
        items = [repr(item) for item in self.iter(order=IterOrder.NO_PARENT)]
        if parent is not None and self._iter_order == IterOrder.PARENT_FIRST:
            items.insert(0, "<parent>")
        if parent is not None and self._iter_order == IterOrder.PARENT_LAST:
            items.append("<parent>")

        ret = "["
        ret += ",".join(items)
        ret += "]"
        return ret



import logging
import tempfile
from collections.abc import Callable
from pathlib import Path
from types import TracebackType
from typing import IO, Any, Protocol

from .autoloader import LoadableProtocol, doLoad
from .timer import PersistentTimer, TimerProtocol


class SavableProtocol(LoadableProtocol, Protocol):
    """
    This protocol lets AutoSaver know how to use a Settings instance without having to
    import the actual Settings class.
    """

    def save(self, file: IO[str], *, blanklines: bool = False) -> None: ...

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None: ...


class AutoSaver:
    """
    AutoSaver is a helper class that can take care of loading and saving settings
    automatically and safely.

    On instantiation, it automatically loads settings from the given file path, if that
    file exists, unless the `load_on_init` argument is set to False.

    When saving, it uses a two-steps mechanism to ensure the atomicity of the operation.
    That is to say, the operation either entirely succeeds or entirely fails; AutoSaver
    will never write an incomplete settings file.

    Saving automatically creates the parent directories of the target path if those
    don't exist yet.

    If the `save_on_update` argument is set to True, AutoSaver will save the settings
    whenever they are updated from inside the application. If the `save_on_delete`
    argument is set to True, AutoSaver will save the settings when its own instance is
    about to get deleted.

    AutoSaver can also be used as a context manager. Using it as a context manager is
    the recommended usage.

    Args:
        settings: The Settings instance to load to and save from.

        path: The full path to the file from which to load and save the settings. If
            this file does not exist yet, it will be created when saving for the first
            time.

        save_on_update: Whether to save the settings when they are updated in any way.
            Default: True.

        save_on_delete: Whether to save the settings when this AutoSaver instance is
            garbage collected. Default: True.

        load_on_init: Whether to load the settings when instantiating this AutoSaver,
            provided the settings path exists. Default: True.

        save_delay: How long to wait, in seconds, before actually saving the settings
            when `save_on_update` is True and an update occurs. Setting this to a few
            seconds will batch updates for that long before triggering a save. If set to
            0, the save is triggered immediately. Default: 0.

        raise_on_error: Whether OS errors occurring while loading and saving the
            settings should raise an exception. If False, errors will only be logged.
            Default: False.

        logger: A logger instance that will be used to log OS errors, if any, while
            loading or saving settings. If none is given, the default root logger will
            be used.

    Example:

    >>> from sunset import AutoSaver, Settings
    >>> class ExampleSettings(Settings):
    ...     ...
    >>> settings = ExampleSettings()
    >>> with AutoSaver(settings, "~/.config/my_app.conf"):  # doctest: +SKIP
    ...     main_program_loop(settings)
    """

    _DIR_MODE: int = 0o755
    _FILE_MODE: int = 0o644
    _ENCODING: str = "UTF-8"

    _path: Path
    _settings: SavableProtocol
    _dirty: bool
    _save_on_update: bool
    _save_on_delete: bool
    _save_delay: float
    _save_timer: TimerProtocol
    _raise_on_error: bool
    _logger: logging.Logger

    def __init__(  # noqa: PLR0913
        self,
        settings: SavableProtocol,
        path: Path | str,
        *,
        save_on_update: bool = True,
        save_on_delete: bool = True,
        load_on_init: bool = True,
        save_delay: float = 0.0,
        raise_on_error: bool = False,
        logger: logging.Logger | None = None,
    ) -> None:
        self._dirty = False
        self._path = Path(path).expanduser()
        self._save_on_update = save_on_update
        self._save_on_delete = save_on_delete
        self._save_delay = save_delay
        self._raise_on_error = raise_on_error
        self._save_timer = PersistentTimer()
        self._settings = settings
        self._settings.onUpdateCall(self._onSettingsUpdated)
        self._logger = logger or logging.getLogger()

        if load_on_init:
            self.doLoad()

    def path(self) -> Path:
        """
        Returns the settings path for this AutoSaver.

        Returns:
            A path. It may or may not exist; that's not up to AutoSaver.
        """

        return self._path

    def doLoad(self) -> bool:
        """
        Load the settings from this AutoSaver's settings file path, if it exists.

        Unsaved settings, if any, will be lost.

        OS errors occurring while loading, if any, will be logged to the logger provided
        to this AutoLoader's constructor.

        If this AutoSaver was constructed with the parameter `raise_on_error` set to
        True, these OS errors will then be re-raised.

        Note that a missing file is not considered an error.

        Returns:
            True if loading succeeded, else False.

        Note:
            Prefer letting AutoSaver load the settings during its initialization in
            order to avoid race conditions.
        """

        return doLoad(
            self._path,
            self._settings,
            raise_on_error=self._raise_on_error,
            logger=self._logger,
            encoding=self._ENCODING,
        )

    def doSave(self) -> bool:
        """
        Unconditionally saves the settings attached to this AutoSaver instance.

        This method uses a two-steps mechanism to perform the save, in order to make the
        save atomic. The settings are first saved to a temporary file, and if
        successful, that temporary file then replaces the actual settings file.

        If the directory where the settings file is located does not exist, this method
        automatically creates it.

        OS errors occurring while saving, if any, will be logged to the logger provided
        to this AutoLoader's constructor.

        If this AutoSaver was constructed with the parameter `raise_on_error` set to
        True, these OS errors will then be re-raised.

        Returns:
            True if saving succeeded, else False.
        """

        self._save_timer.cancel()

        directory = self._path.parent
        directory.mkdir(parents=True, mode=self._DIR_MODE, exist_ok=True)

        try:
            with tempfile.NamedTemporaryFile(
                dir=directory,
                prefix=self._path.name,
                mode="xt",
                encoding=self._ENCODING,
                delete=False,
            ) as tmp:
                self._logger.debug("Saving settings file '%s'...", self._path)
                self._settings.save(tmp.file, blanklines=True)
                tmp_path = Path(tmp.name)
                tmp_path.chmod(self._FILE_MODE)
                tmp_path.rename(self._path)
                self._logger.debug("Saved.")

        except OSError as e:
            msg = f"Failed to save settings to {self._path}: {e}"
            self._logger.error(msg)  # noqa: TRY400

            if self._raise_on_error:
                raise

            return False

        self._dirty = False
        return True

    def saveIfNeeded(self) -> bool:
        """
        Performs a save if and only if there are pending, unsaved changes in the
        settings attached to this AutoSaver instance.

        Returns:
            True if a save was performed and succeeded, else False.
        """

        save_needed = self._dirty

        self._save_timer.cancel()

        return self.doSave() if save_needed else False

    def _onSettingsUpdated(self, _: Any) -> None:  # noqa: ANN401
        self._dirty = True
        if self._save_on_update:
            self._save_timer.start(self._save_delay, self.saveIfNeeded)

    def __del__(self) -> None:
        if self._save_on_delete:
            self.saveIfNeeded()

    def __enter__(self) -> "AutoSaver":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.saveIfNeeded()



import threading
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from functools import wraps
from typing import ParamSpec, TypeVar

_R = TypeVar("_R")
_P = ParamSpec("_P")


class RWLock:
    """
    A reentrant read-write lock.

    Any number of threads can hold the lock for reads concurrently. Only one thread at a
    time can hold the lock for writes. When a thread holds the lock for writes, no other
    thread can hold the lock for reads, and vice versa.

    The lock is reentrant over both reads and writes in any combination.
    """

    _readers: dict[int, int]
    _readers_update: threading.Condition
    _writer_lock: threading.RLock

    def __init__(self) -> None:
        self._readers = {}
        self._readers_update = threading.Condition()
        self._writer_lock = threading.RLock()

    @staticmethod
    def _current_thread_id() -> int:
        return threading.current_thread().ident or 0

    def _acquire_read_lock(self, *, blocking: bool = True) -> bool:
        # Grab the writer lock before adding a reader, so that we don't add more readers
        # when there's a writer waiting.
        if not self._writer_lock.acquire(blocking):
            return False

        try:
            with self._readers_update:
                thread_id = self._current_thread_id()
                self._readers.setdefault(thread_id, 0)
                self._readers[thread_id] += 1

                # Enforce critical invariant.
                assert all(count > 0 for count in self._readers.values())  # noqa: S101

        finally:
            self._writer_lock.release()

        return True

    def _release_read_lock(self) -> None:
        with self._readers_update:
            thread_id = self._current_thread_id()
            if thread_id not in self._readers:
                msg = "Attempted to release read lock, but it was not held"
                raise RuntimeError(msg)

            self._readers[thread_id] -= 1

            if self._readers[thread_id] == 0:
                del self._readers[thread_id]
                self._readers_update.notify()

            # Enforce critical invariant.
            assert all(count > 0 for count in self._readers.values())  # noqa: S101

    def _acquire_write_lock(self, *, blocking: bool = True) -> bool:
        if not self._writer_lock.acquire(blocking):
            return False

        with self._readers_update:
            while self._readers:
                if (
                    len(self._readers) == 1
                    and self._current_thread_id() in self._readers
                ):
                    # Don't actually block if the only active reader is the current
                    # thread. This allows RWLock to be reentrant on any combination of
                    # reads and writes.
                    break

                if not blocking:
                    self._writer_lock.release()
                    return False

                self._readers_update.wait()

        return True

    def _release_write_lock(self) -> None:
        self._writer_lock.release()

    @contextmanager
    def lock_reads(self) -> Iterator[None]:
        self._acquire_read_lock(blocking=True)
        yield
        self._release_read_lock()

    @contextmanager
    def lock_writes(self) -> Iterator[None]:
        self._acquire_write_lock(blocking=True)
        yield
        self._release_write_lock()

    def with_read_lock(self, func: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(func)
        def locked_func(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            with self.lock_reads():
                return func(*args, **kwargs)

        return locked_func

    def with_write_lock(self, func: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(func)
        def locked_func(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            with self.lock_writes():
                return func(*args, **kwargs)

        return locked_func


SettingsLock = RWLock()



import threading
from collections.abc import Callable
from typing import Any, Protocol


class TimerProtocol(Protocol):
    def __init__(self, *, looping: bool = False) -> None: ...

    def start(self, interval: float, function: Callable[[], Any]) -> None: ...

    def cancel(self) -> None: ...


class PersistentTimer:
    _timer: threading.Timer | None
    _lock: threading.Lock
    _looping: bool

    def __init__(self, *, looping: bool = False) -> None:
        self._timer = None
        self._looping = looping
        self._lock = threading.Lock()

    def start(self, interval: float, function: Callable[[], Any]) -> None:
        if interval <= 0.0:
            self._triggerFunction(function)
            return

        with self._lock:
            if self._timer is None:
                self._timer = threading.Timer(
                    interval, self._triggerFunction, [function]
                )
                self._timer.start()

    def cancel(self) -> None:
        with self._lock:
            if self._timer is not None:
                self._timer.cancel()
                self._timer = None

    def _triggerFunction(self, function: Callable[[], Any]) -> None:
        interval = self._timer.interval if self._timer else None
        self.cancel()
        function()

        if self._looping and interval is not None:
            self.start(interval, function)

    def __del__(self) -> None:
        self.cancel()



"SunsetSettings: a type-safe, extensible INI-style settings system with inheritance."

__project__ = "SunsetSettings"
__version__ = "0.7.0"
__author__ = "P. Varet"
__copyright__ = "2022-2025, P. Varet"

import warnings

from sunset import exporter, serializers, sets, stringutils
from sunset.autoloader import AutoLoader
from sunset.autosaver import AutoSaver
from sunset.bunch import Bunch
from sunset.enum_serializer import SerializableEnum, SerializableFlag
from sunset.key import Key
from sunset.list import List
from sunset.protocols import Serializable, Serializer
from sunset.settings import Settings, normalize
from sunset.timer import PersistentTimer


class Bundle(Bunch):
    # Backward compatibility: Bunch used to be called Bundle prior to version 0.4.0.
    # Retain compatibility by keeping this name around for a few versions.
    def __init__(self) -> None:
        version_tuple = tuple(
            int(s) if s.isdigit() else 0 for s in __version__.split(".")
        )
        msg = "'Bundle' is deprecated. Use 'Bunch' instead."
        if version_tuple < (1, 0):
            warnings.warn(msg, DeprecationWarning, stacklevel=1)
        else:  # pragma: no cover
            raise DeprecationWarning(msg)
        super().__init__()


__all__ = [
    "AutoLoader",
    "AutoSaver",
    "Bunch",
    "Bundle",
    "Key",
    "List",
    "PersistentTimer",
    "Serializable",
    "SerializableEnum",
    "SerializableFlag",
    "Serializer",
    "Settings",
    "exporter",
    "normalize",
    "serializers",
    "sets",
    "stringutils",
]



import enum
from typing import Generic, TypeVar

from sunset.enum_serializer import EnumSerializer
from sunset.protocols import Serializable, Serializer

_T = TypeVar("_T")
_Serializable = TypeVar("_Serializable", bound=Serializable)
_Castable = TypeVar("_Castable", int, float, str)
_Bool = TypeVar("_Bool", bound=bool)


class StraightCastSerializer(Generic[_Castable]):
    _type: type[_Castable]

    def __init__(self, type_: type[_Castable]) -> None:
        self._type = type_

    def toStr(self, value: _Castable) -> str:
        return str(value)

    def fromStr(self, string: str) -> _Castable | None:
        try:
            return self._type(string)
        except ValueError:
            return None


class BoolSerializer(Generic[_Bool]):
    def __init__(self, type_: type[_Bool]) -> None:
        self._type = type_

    def toStr(self, value: _Bool) -> str:
        return "true" if value else "false"

    def fromStr(self, string: str) -> _Bool | None:
        string = string.strip().lower()
        if string in ("true", "yes", "y", "1"):
            return self._type(True)  # noqa: FBT003
        if string in ("false", "no", "n", "0"):
            return self._type(False)  # noqa: FBT003
        return None


class SerializableSerializer(Generic[_Serializable]):
    _type: type[_Serializable]

    def __init__(self, type_: type[_Serializable]) -> None:
        self._type = type_

    def toStr(self, value: _Serializable) -> str:
        return value.toStr()

    def fromStr(self, string: str) -> _Serializable | None:
        return self._type.fromStr(string)


def lookup(type_: type[_T]) -> Serializer[_T] | None:
    if issubclass(type_, Serializable):
        return SerializableSerializer(type_)

    if issubclass(type_, bool):
        return BoolSerializer(type_)

    if issubclass(type_, enum.Enum):
        return EnumSerializer(type_)

    # Note: these need to come after the Enum case, because Enum can be a
    # subclass of these.

    if issubclass(type_, int) or issubclass(type_, float) or issubclass(type_, str):
        return StraightCastSerializer(type_)

    return None



from collections.abc import Callable

import pytest

from sunset import sets


class TestWeakCallableSet:
    def test_function(self) -> None:
        def test(_: str) -> None:
            pass

        reg: sets.WeakCallableSet[Callable[[str], None]] = sets.WeakCallableSet()

        assert len(reg) == 0

        reg.add(test)
        assert len(reg) == 1
        assert test in reg

        del test
        assert len(reg) == 0

    def test_method(self) -> None:
        class Test:
            def test(self, _: str) -> None:
                pass

        reg: sets.WeakCallableSet[Callable[[str], None]] = sets.WeakCallableSet()

        t = Test()

        assert len(reg) == 0
        assert t.test not in reg

        reg.add(t.test)
        assert len(reg) == 1
        assert t.test in reg

        del t
        assert len(reg) == 0

    def test_discard_function(self) -> None:
        def test(_: int) -> None:
            pass

        reg: sets.WeakCallableSet[Callable[[int], None]] = sets.WeakCallableSet()

        reg.add(test)
        assert test in reg
        reg.discard(test)
        assert test not in reg

    def test_discard_method(self) -> None:
        class Test:
            def test(self, _: int) -> None:
                pass

        t = Test()

        reg: sets.WeakCallableSet[Callable[[int], None]] = sets.WeakCallableSet()

        reg.add(t.test)
        assert t.test in reg
        reg.discard(t.test)
        assert t.test not in reg

    def test_pop(self) -> None:
        def test(_: bool) -> None:
            pass

        reg: sets.WeakCallableSet[Callable[[bool], None]] = sets.WeakCallableSet()

        reg.add(test)
        assert test in reg
        other = reg.pop()
        assert test not in reg
        assert other is test

    def test_function_added_once(self) -> None:
        def test(_: str) -> None:
            pass

        reg: sets.WeakCallableSet[Callable[[str], None]] = sets.WeakCallableSet()

        reg.add(test)
        assert len(reg) == 1
        reg.add(test)
        assert len(reg) == 1

    def test_same_method_added_once(self) -> None:
        class Test:
            def test(self, _: str) -> None:
                pass

        reg: sets.WeakCallableSet[Callable[[str], None]] = sets.WeakCallableSet()

        t = Test()

        reg.add(t.test)
        assert len(reg) == 1
        reg.add(t.test)
        assert len(reg) == 1

    def test_non_hashable_element(self) -> None:
        class Test(list[int]):
            def test(self, _: int) -> None:
                pass

        t = Test()

        with pytest.raises(TypeError):
            # Prove that type Test is not hashable.

            hash(t)

        reg: sets.WeakCallableSet[Callable[[int], None]] = sets.WeakCallableSet()

        reg.add(t.test)
        assert t.test in reg

    def test_different_method_added_twice(self) -> None:
        class Test:
            def test(self, _: str) -> None:
                pass

        reg: sets.WeakCallableSet[Callable[[str], None]] = sets.WeakCallableSet()

        t1 = Test()
        t2 = Test()

        reg.add(t1.test)
        assert len(reg) == 1
        reg.add(t2.test)
        assert len(reg) == 2



from __future__ import annotations

import logging
from pathlib import Path
from typing import IO, TYPE_CHECKING, Protocol

from .timer import PersistentTimer

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Callable
    from types import TracebackType
    from typing import Any, Self

    from sunset import Settings


class LoadableProtocol(Protocol):
    """
    This protocol lets AutoLoader know how to use a Settings instance without
    having to import the actual Settings class.
    """

    def load(self, file: IO[str]) -> None: ...


class MonitorProtocol(Protocol):
    def __init__(
        self, path: Path, callback: Callable[[], Any], monitor_period_s: int = 1
    ) -> None: ...

    def start(self) -> None: ...

    def stop(self) -> None: ...

    def triggerCallback(self) -> None: ...


class MonitorForChange:
    """
    A simple file monitor that calls a callback function when the file changes.
    """

    _path: Path
    _callback: Callable[[], Any]
    _monitor_period_s: int
    _timer: PersistentTimer | None = None

    def __init__(
        self,
        path: Path,
        callback: Callable[[], Any],
        monitor_period_s: int = 1,
    ) -> None:
        self._path = path
        self._callback = callback
        self._monitor_period_s = monitor_period_s
        self._timer = None

    def start(self) -> None:
        if self._timer is not None:
            return

        try:
            last_modified = self._path.stat().st_mtime
            pending_update = False
        except OSError:
            last_modified = None
            pending_update = True

        def loop() -> None:
            nonlocal last_modified, pending_update
            try:
                new_last_modified = self._path.stat().st_mtime
            except OSError:
                last_modified = None
                pending_update = True
            else:
                if new_last_modified == last_modified:
                    if pending_update:
                        self.triggerCallback()
                        pending_update = False
                else:
                    pending_update = True
                last_modified = new_last_modified

        self._timer = PersistentTimer(looping=True)
        self._timer.start(self._monitor_period_s, loop)

    def stop(self) -> None:
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

    def triggerCallback(self) -> None:
        self._callback()


def doLoad(
    path: Path,
    settings: LoadableProtocol,
    *,
    raise_on_error: bool = False,
    logger: logging.Logger | None = None,
    encoding: str = "UTF-8",
) -> bool:
    logger = logger or logging.getLogger()

    # Deal gracefully with missing paths. That will be a common case when the file
    # has not been saved yet.
    if not path.exists():
        msg = f"Not loading settings -- path {path} does not exist."
        logger.info(msg)
        return False

    msg = f"Loading settings from {path}..."
    logger.debug(msg)

    try:
        with path.open(encoding=encoding) as f:
            settings.load(f)

    except OSError as e:
        msg = f"Failed to load settings from {path}: {e}"
        logger.error(msg)  # noqa: TRY400

        if raise_on_error:
            raise

        return False

    logger.debug("Loaded.")

    return True


class AutoLoader:
    """
    A helper that loads settings from the given file and reloads them when the file
    changes.

    On instantiation, it automatically loads settings from the given file path.

    Afterward, it periodically checks the file for changes and reloads the settings
    whenever the file is modified. Note that the settings are only reloaded when the
    file has not been modified for a few seconds, in order to avoid attempting to load a
    file that is still being written to.

    AutoLoader can also be used as a context manager. Using it as a context manager is
    the recommended usage.

    Args:
        settings: The Settings instance to load settings for.

        path: The full path to the file to load the settings from.

        raise_on_error: Whether OS errors occurring while loading the settings should
            raise an exception. If False, errors will only be logged. Default: False.

        logger: A logger instance that will be used to log OS errors, if any, while
            loading settings. If none is given, the default root logger will be used.

    Example:

    >>> from sunset import Settings, AutoLoader
    >>> class ExampleSettings(Settings):
    ...     ...
    >>> settings = ExampleSettings()
    >>> with AutoLoader(settings, "~/.config/my_app.conf"):  # doctest: +SKIP
    ...     main_program_loop(settings)

    """

    _ENCODING: str = "UTF-8"

    _settings: Settings
    _path: Path
    _raise_on_error: bool
    _logger: logging.Logger

    def __init__(
        self,
        settings: Settings,
        path: str | Path,
        *,
        raise_on_error: bool = False,
        logger: logging.Logger | None = None,
        _monitor_factory: Callable[[Any, Any], MonitorProtocol] = MonitorForChange,
    ) -> None:
        self._settings = settings
        self._path = Path(path).expanduser()
        self._raise_on_error = raise_on_error
        self._monitor = _monitor_factory(self._path, self.doLoad)
        self._logger = logger or logging.getLogger()

        self.doLoad()
        self._monitor.start()

    def doLoad(self) -> bool:
        """
        Perform a load now, without waiting for the next detected change in the settings
        file.

        Unsaved settings, if any, will be lost.

        OS errors occurring during the load, if any, will be logged to the logger
        provided to this AutoLoader's constructor.

        If this AutoLoader was constructed with the parameter `raise_on_error` set to
        True, these OS errors will then be re-raised.

        Note that a missing file is not considered an error.

        Returns:
            True if loading succeeded, else False.

        """
        return doLoad(
            self._path, self._settings, logger=self._logger, encoding=self._ENCODING
        )

    def path(self) -> Path:
        return self._path

    def __enter__(self) -> Self:
        self._monitor.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._monitor.stop()

    def __del__(self) -> None:
        self._monitor.stop()



from typing import Any

import pytest

from sunset.stringutils import collate_by_prefix, split_on


@pytest.mark.parametrize(
    "string, expected",
    [
        ("", ("", "")),
        ("a", ("a", "")),
        ("a.b", ("a", "b")),
        ("a.b.c", ("a", "b.c")),
        (".a", ("", "a")),
        ("..a", ("", ".a")),
        ("a.", ("a", "")),
        (".a.", ("", "a.")),
    ],
)
def test_split_on(string: str, expected: tuple[str, str]) -> None:
    assert split_on(".")(string) == expected


_SENTINEL = "test"


@pytest.mark.parametrize(
    "inputs, expected",
    [
        ([], {}),
        ([("", _SENTINEL), ("", _SENTINEL)], {"": [("", _SENTINEL), ("", _SENTINEL)]}),
        (
            [("a", _SENTINEL), ("b", _SENTINEL)],
            {"a": [("", _SENTINEL)], "b": [("", _SENTINEL)]},
        ),
        (
            [("a.b", _SENTINEL), ("a.c", _SENTINEL)],
            {"a": [("b", _SENTINEL), ("c", _SENTINEL)]},
        ),
        (
            [("a.b.c", _SENTINEL), ("a.b.d", _SENTINEL), ("x.b.c", _SENTINEL)],
            {
                "a": [("b.c", _SENTINEL), ("b.d", _SENTINEL)],
                "x": [("b.c", _SENTINEL)],
            },
        ),
        (
            [
                ("a", _SENTINEL),
                ("a.b", _SENTINEL),
                ("a.b.c", _SENTINEL),
                ("b.c", _SENTINEL),
            ],
            {
                "a": [("", _SENTINEL), ("b", _SENTINEL), ("b.c", _SENTINEL)],
                "b": [("c", _SENTINEL)],
            },
        ),
    ],
)
def test_collate_by_prefix(
    inputs: list[tuple[str, Any]], expected: dict[str, list[tuple[str, Any]]]
) -> None:
    splitter = split_on(".")

    assert collate_by_prefix(inputs, splitter=splitter) == expected



from collections.abc import MutableSet
from typing import Any

from sunset import sets


class Dummy:
    pass


class TestNonHashableSet:
    def test_mutableset(self) -> None:
        s: MutableSet[Any] = sets.NonHashableSet()
        item = Dummy()

        assert item not in s
        assert len(s) == 0
        assert list(s) == []

        s.add(item)
        s.add(item)
        assert item in s
        assert len(s) == 1
        assert list(s) == [item]

        s.discard(item)
        assert item not in s
        assert len(s) == 0
        assert list(s) == []

    def test_persistence(self) -> None:
        s: MutableSet[Any] = sets.NonHashableSet()
        item = Dummy()

        s.add(item)
        assert len(s) == 1

        del item
        assert len(s) == 1


class TestWeakNonHashableSet:
    def test_mutableset(self) -> None:
        s: MutableSet[Any] = sets.WeakNonHashableSet()
        item = Dummy()

        assert item not in s
        assert len(s) == 0
        assert list(s) == []

        s.add(item)
        s.add(item)
        assert item in s
        assert len(s) == 1
        assert list(s) == [item]

        s.discard(item)
        assert item not in s
        assert len(s) == 0
        assert list(s) == []

    def test_persistence(self) -> None:
        s: MutableSet[Any] = sets.WeakNonHashableSet()
        item = Dummy()

        s.add(item)
        assert len(s) == 1

        del item
        assert len(s) == 0



import tempfile
import time

from sunset import Key, Settings, AutoLoader


class DemoSettings(Settings):
    a: Key[int] = Key(default=-1)


if __name__ == "__main__":

    def report_value_changed(value: int) -> None:
        print(f"    -> Setting value changed: a={value}")
        print()

    settings = DemoSettings()
    settings.a.onValueChangeCall(report_value_changed)

    with tempfile.NamedTemporaryFile(delete=False, mode="w+", encoding="utf-8") as file:
        with AutoLoader(settings, file.name) as loader:
            for i in range(1, 11):
                time.sleep(0.5)

                contents = f"[main]\na = {i}\n"
                print("Writing new settings file:")
                print()
                print(contents)

                file.seek(0)
                file.write(contents)
                file.truncate()
                file.flush()

                # The autoloader should detect that the file was modified, and trigger a
                # load in the background. When it does, the value of settings.a will be
                # updated, and the callback triggered.

                time.sleep(2.5)



from pytest_mock import MockerFixture

from sunset.notifier import Notifier


class TestNotifier:
    def test_function_is_called(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        notifier = Notifier[str, int]()

        notifier.add(callback)

        callback.assert_not_called()
        notifier.trigger("test", 12)
        callback.assert_called_once_with("test", 12)

    def test_function_added_multiple_times_is_called_once(
        self, mocker: MockerFixture
    ) -> None:
        notifier = Notifier[str]()

        callback = mocker.stub()

        notifier.add(callback)
        notifier.add(callback)
        notifier.add(callback)

        notifier.trigger("test")

        callback.assert_called_once_with("test")

    def test_all_added_functions_called(self, mocker: MockerFixture) -> None:
        notifier = Notifier[str]()

        callbacks = [mocker.stub() for _ in range(10)]

        for callback in callbacks:
            notifier.add(callback)

        notifier.trigger("test")

        for callback in callbacks:
            callback.assert_called_once_with("test")

    def test_discarded_function_is_no_longer_called(
        self, mocker: MockerFixture
    ) -> None:
        notifier = Notifier[str]()

        callback1 = mocker.stub()
        callback2 = mocker.stub()

        notifier.add(callback1)
        notifier.add(callback2)
        notifier.discard(callback1)

        notifier.trigger("test")

        callback1.assert_not_called()
        callback2.assert_called_once_with("test")

    def test_discarding_non_member_is_no_op(self) -> None:
        notifier = Notifier[str]()

        notifier.discard(lambda _: None)

    def test_notifier_doesnt_keep_reference_to_function(
        self, mocker: MockerFixture
    ) -> None:
        notifier = Notifier[str]()

        test_list: list[str] = []

        def test(value: str) -> None:
            test_list.append(value)

        notifier.add(test)
        notifier.trigger("test")

        assert test_list == ["test"]

        test_list.clear()
        del test
        notifier.trigger("test")

        assert test_list == []

    def test_inhibit(self, mocker: MockerFixture) -> None:
        notifier = Notifier[str]()

        callback = mocker.stub()

        notifier.add(callback)

        with notifier.inhibit():
            notifier.trigger("test")

        callback.assert_not_called()

        notifier.trigger("test")
        callback.assert_called_once_with("test")

    def test_nested_inhibit(self, mocker: MockerFixture) -> None:
        notifier = Notifier[str]()

        callback = mocker.stub()

        notifier.add(callback)

        with notifier.inhibit():
            with notifier.inhibit():
                notifier.trigger("test")

        callback.assert_not_called()

        notifier.trigger("test")
        callback.assert_called_once_with("test")



import logging
import pathlib

import pytest
from pytest_mock import MockerFixture

from sunset import AutoSaver, Key, Settings

from .test_timer import MockTimer


class ExampleSettings(Settings):
    key_str = Key(default="")


class TestAutosaver:
    def test_load_on_init(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings_file.write_text("[main]\nkey_str = test\n")

        settings = ExampleSettings()

        assert settings.key_str.get() == ""

        AutoSaver(settings, settings_file)

        assert settings.key_str.get() == "test"

    def test_no_load_on_init(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings_file.write_text("[main]\nkey_str = test\n")

        settings = ExampleSettings()

        assert settings.key_str.get() == ""

        AutoSaver(settings, settings_file, load_on_init=False)

        assert settings.key_str.get() == ""

    def test_no_unneeded_saving(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings = ExampleSettings()
        autosaver = AutoSaver(
            settings, settings_file, save_on_update=True, save_on_delete=True
        )

        assert not settings_file.exists()

        autosaver.__del__()  # Simulate garbage collection.

        assert not settings_file.exists()

    def test_autosave_on_update(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings = ExampleSettings()
        autosaver = AutoSaver(settings, settings_file, save_on_delete=False)

        assert not settings_file.exists()

        settings.key_str.set("test")

        assert settings_file.exists()
        assert settings_file.read_text() == "[main]\nkey_str = test\n"

        del autosaver  # So that it's not unused.

    def test_autosave_on_delete(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings = ExampleSettings()
        autosaver = AutoSaver(settings, settings_file, save_on_update=False)

        settings.key_str.set("test")

        assert not settings_file.exists()

        autosaver.__del__()  # Simulate garbage collection.

        assert settings_file.exists()
        assert settings_file.read_text() == "[main]\nkey_str = test\n"

    def test_autosave_creates_missing_dirs(self, tmp_path: pathlib.Path) -> None:
        settings = ExampleSettings()
        settings_file = tmp_path / "multiple" / "levels" / "deep" / "test.conf"
        autosaver = AutoSaver(
            settings, settings_file, save_on_update=False, save_on_delete=False
        )

        assert not settings_file.parent.exists()

        autosaver.doSave()

        assert settings_file.parent.exists()

    def test_context_manager(self, tmp_path: pathlib.Path) -> None:
        settings = ExampleSettings()
        settings_file = tmp_path / "test.conf"
        with AutoSaver(
            settings, settings_file, save_on_update=False, save_on_delete=False
        ):
            settings.key_str.set("test")
            assert not settings_file.exists()

        assert settings_file.exists()
        assert settings_file.read_text() == "[main]\nkey_str = test\n"

    def test_delayed_save(self, tmp_path: pathlib.Path) -> None:
        settings_file = tmp_path / "test.conf"
        settings = ExampleSettings()
        autosaver = AutoSaver(
            settings, settings_file, save_on_delete=False, save_delay=2.0
        )

        mock_timer = MockTimer()
        autosaver._save_timer = mock_timer

        assert not settings_file.exists()

        settings.key_str.set("test 1")

        assert not settings_file.exists()

        settings.key_str.set("test 2")
        settings.key_str.set("test 3")

        mock_timer.advanceTime(1.0)

        assert not settings_file.exists()

        mock_timer.advanceTime(2.0)

        assert settings_file.exists()

    def test_expand_user(self, monkeypatch: pytest.MonkeyPatch) -> None:
        def expanduser(path: pathlib.Path) -> pathlib.Path:
            if not str(path).startswith("~"):
                return path
            return pathlib.Path("HOME") / str(path).lstrip("~").lstrip("/")

        monkeypatch.setattr(pathlib.Path, "expanduser", expanduser)

        saver1 = AutoSaver(ExampleSettings(), "/no/tilde", load_on_init=False)
        assert str(saver1.path()) == "/no/tilde"

        saver2 = AutoSaver(ExampleSettings(), "~/with/tilde", load_on_init=False)
        assert str(saver2.path()) == "HOME/with/tilde"

    def test_no_exception_raised_if_logger_provided(
        self, tmp_path: pathlib.Path, mocker: MockerFixture
    ) -> None:
        settings_file = tmp_path / "actually_a_folder"
        settings_file.mkdir(parents=True)

        logger_stub = mocker.MagicMock(logging.Logger)

        saver = AutoSaver(
            ExampleSettings(),
            settings_file,
            load_on_init=False,
            logger=logger_stub,
        )

        logger_stub.error.assert_not_called()
        assert settings_file.exists()
        assert not saver.doLoad()
        logger_stub.error.assert_called_once()
        logger_stub.error.reset_mock()

        assert not saver.doSave()
        logger_stub.error.assert_called_once()

    def test_exception_raised_if_raise_on_exception(
        self, tmp_path: pathlib.Path, mocker: MockerFixture
    ) -> None:
        settings_file = tmp_path / "actually_a_folder"
        settings_file.mkdir(parents=True)

        logger_stub = mocker.MagicMock(logging.Logger)

        saver = AutoSaver(
            ExampleSettings(),
            settings_file,
            load_on_init=False,
            raise_on_error=True,
            logger=logger_stub,
        )

        assert settings_file.exists()
        with pytest.raises(OSError):
            saver.doLoad()



from typing import Optional

import pytest
from pytest_mock import MockerFixture

from sunset import Bunch, Key, SerializableEnum, SerializableFlag, protocols


class ExampleSerializable:
    def __init__(self, value: str) -> None:
        self._value = value

    def toStr(self) -> str:
        return self._value

    @staticmethod
    def fromStr(string: str) -> Optional["ExampleSerializable"]:
        return ExampleSerializable(string)


class ExampleEnum(SerializableEnum):
    ONE = 1
    TWO = 2


class ExampleFlag(SerializableFlag):
    ONE = 1
    TWO = 2
    THREE = ONE | TWO


class TestKey:
    def test_protocol_implementation(self) -> None:
        key = Key(default="")
        assert isinstance(key, protocols.Field)

    def test_default(self) -> None:
        assert Key(default="default").get() == "default"
        assert Key(default=0).get() == 0
        assert type(Key(default=False).get()) is bool
        assert not Key(default=False).get()
        assert Key(default=12.345e-67).get() == 12.345e-67
        assert Key(default=ExampleEnum.ONE).get() == ExampleEnum.ONE
        assert Key(default=ExampleFlag.ONE | ExampleFlag.TWO).get() == ExampleFlag.THREE
        with pytest.raises(TypeError):
            Key(default=object())

    def test_set(self) -> None:
        key = Key(default="test")
        key.set("other")
        assert key.get() == "other"

    def test_clear(self) -> None:
        key = Key(default="default")
        key.set("other")
        assert key.get() != "default"

        key.clear()
        assert key.get() == "default"

    def test_fallback(self) -> None:
        key = Key(default=0)
        child_key = Key(default=0)
        child_key.setParent(key)

        assert key.fallback() == 0
        assert child_key.fallback() == 0

        child_key.set(1)

        assert key.fallback() == 0
        assert child_key.fallback() == 0

        key.set(2)

        assert key.fallback() == 0
        assert child_key.fallback() == 2

    def test_serializable_type(self) -> None:
        value = ExampleSerializable.fromStr("dummy")
        assert value is not None
        key: Key[ExampleSerializable] = Key(default=value)

        assert key.get().toStr() == "dummy"

    def test_missing_serializer(self) -> None:
        class NotSerializable:
            pass

        with pytest.raises(TypeError):
            Key(default=NotSerializable())

    def test_explicit_serializer(self) -> None:
        class NotSerializable:
            pass

        class Serializer:
            def toStr(self, value: NotSerializable) -> str:
                return "test"

            def fromStr(self, string: str) -> NotSerializable:
                return NotSerializable()

        key = Key(default=NotSerializable(), serializer=Serializer())
        key.set(NotSerializable())
        assert list(key.dumpFields()) == [("", "test")]
        assert key.restoreFields([("", "")])

    def test_serializer_override(self) -> None:
        class Serializer:
            def toStr(self, value: str) -> str:
                return value + "-TEST"

            def fromStr(self, string: str) -> str | None:
                if string.endswith("-TEST"):
                    return string[:-5]
                return None

        key: Key[str] = Key(default="", serializer=Serializer())
        key.set("value")
        assert list(key.dumpFields()) == [("", "value-TEST")]

        assert key.restoreFields([("", "other value-TEST")])
        assert key.get() == "other value"

        assert not key.restoreFields([("", "(invalid)")])
        assert key.get() == "other value"

    def test_validator(self) -> None:
        def isEven(i: int) -> bool:
            return i % 2 == 0

        key: Key[int] = Key(default=0, validator=isEven)

        assert not key.set(1)
        assert key.get() == 0

        assert key.set(2)
        assert key.get() == 2

        assert not key.restoreFields([("", "3")])
        assert key.get() == 2

        assert key.restoreFields([("", "4")])
        assert key.get() == 4

        def isOdd(i: int) -> bool:
            return i % 2 == 1

        key.setValidator(isOdd)

        assert key.set(5)
        assert key.get() == 5

        assert not key.set(6)
        assert key.get() == 5

    def test_value_change_callback(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        key = Key(default="default")

        key.onValueChangeCall(callback)

        key.set("default")
        callback.assert_not_called()
        callback.reset_mock()

        key.set("not default")
        callback.assert_called_once_with("not default")
        callback.reset_mock()

        key.clear()
        callback.assert_called_once_with("default")
        callback.reset_mock()

        key.clear()
        callback.assert_not_called()

    def test_value_change_callback_inheritance(self, mocker: MockerFixture) -> None:
        callback_sub_child1 = mocker.stub()
        callback_sub_child2 = mocker.stub()

        # Set up keys so that one parent key has two children key, each with one
        # child of its own.

        parent_key = Key(default="default")
        child1_key = Key(default="default")
        child2_key = Key(default="default")
        sub_child1_key = Key(default="default")
        sub_child2_key = Key(default="default")

        child1_key.setParent(parent_key)
        child2_key.setParent(parent_key)
        sub_child1_key.setParent(child1_key)
        sub_child2_key.setParent(child2_key)

        # Set up callbacks on the grandchildren keys.

        sub_child1_key.onValueChangeCall(callback_sub_child1)
        sub_child2_key.onValueChangeCall(callback_sub_child2)

        # One of the first generation children gets a value set on it. That
        # should count as set even if the value is the same as the default value
        # of the parent! But the sub-child's apparent value does not change, so
        # its callback shouldn't be called.

        child1_key.set("default")
        callback_sub_child1.assert_not_called()

        # If the toplevel parent's value changes to something new, the apparent
        # value change should be propagated down the inheritance chain through
        # all *unset* children. So only the second one in this case.

        parent_key.set("test")
        callback_sub_child1.assert_not_called()
        callback_sub_child2.assert_called_once_with("test")
        callback_sub_child2.reset_mock()

        # Updating the intermediate child should update the value of its own
        # child.

        child1_key.set("test")
        callback_sub_child1.assert_called_once_with("test")
        callback_sub_child1.reset_mock()

        # Clearing it causes its child to now inherit the value from the
        # toplevel parent, but that value is the same; so the child's value is
        # not updated.

        child1_key.clear()
        callback_sub_child1.assert_not_called()
        callback_sub_child1.reset_mock()

        # Clearing the parent does change the value inherited by the grandchild.

        parent_key.clear()
        callback_sub_child1.assert_called_once_with("default")
        callback_sub_child1.reset_mock()

    def test_key_updated_callback(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        key = Key(default="default")
        assert isinstance(key, protocols.UpdateNotifier)

        key.onUpdateCall(callback)

        key.set("default")
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.set("not default")
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.set("not default")
        callback.assert_not_called()
        callback.reset_mock()

        key.clear()
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.clear()
        callback.assert_not_called()
        callback.reset_mock()

    def test_callback_type_is_flexible(self) -> None:
        key = Key("")

        class Dummy:
            pass

        def callback1(_: Key[str]) -> Dummy: ...

        def callback2(_: str) -> Dummy: ...

        key.onUpdateCall(callback1)
        key.onValueChangeCall(callback2)

    def test_updater(self) -> None:
        key = Key("")

        def updater(value: str) -> str:
            return value + "x"

        assert key.get() == ""
        key.updateValue(updater)
        assert key.get() == "x"
        key.updateValue(updater)
        assert key.get() == "xx"

    def test_inheritance(self) -> None:
        parent_key = Key(default="default a")
        child_key = Key(default="")

        assert child_key not in parent_key.children()
        child_key.setParent(parent_key)
        assert child_key in parent_key.children()

        assert child_key.get() == "default a"

        parent_key.set("new a")
        assert child_key.get() == "new a"

        child_key.set("new b")
        assert child_key.get() == "new b"

        parent_key.clear()
        assert child_key.get() == "new b"

        child_key.clear()
        assert child_key.get() == "default a"

    def test_inherit_wrong_type(self) -> None:
        parent_key = Key(default="str")
        child_key = Key(default=0)

        # Ignore the type error, as it's the whole point of the test.

        child_key.setParent(parent_key)  # type: ignore

        assert child_key.parent() is None

    def test_inherit_revert(self) -> None:
        parent_key = Key(default="default a")
        child_key = Key(default="default b")

        assert child_key not in parent_key.children()
        assert child_key.parent() is None

        child_key.setParent(parent_key)

        assert child_key in parent_key.children()
        assert child_key.parent() is parent_key

        child_key.setParent(None)

        assert child_key not in parent_key.children()
        assert child_key.parent() is None

    def test_repr(self) -> None:
        key1 = Key(default="test")
        key2 = Key(default=12)
        key3 = Key(default="  test\ntest  ")

        assert repr(key1) == "<Key[str]:(test)>"
        assert repr(key2) == "<Key[int]:(12)>"
        assert repr(key3) == '<Key[str]:("  test\\ntest  ")>'

        key1.set("test")
        key2.set(12)
        key3.set("  test\ntest  ")

        assert repr(key1) == "<Key[str]:test>"
        assert repr(key2) == "<Key[int]:12>"
        assert repr(key3) == '<Key[str]:"  test\\ntest  ">'

    def test_reparenting(self) -> None:
        key1 = Key(default="default a")
        key2 = Key(default="default b")
        child_key = Key(default="default c")

        assert child_key not in key1.children()
        assert child_key not in key2.children()

        child_key.setParent(key1)
        assert child_key in key1.children()
        assert child_key not in key2.children()

        child_key.setParent(key2)
        assert child_key not in key1.children()
        assert child_key in key2.children()

        child_key.setParent(None)
        assert child_key not in key1.children()
        assert child_key not in key2.children()

    def test_callback_triggered_on_parent_value_change(
        self, mocker: MockerFixture
    ) -> None:
        stub = mocker.stub()

        parent_key = Key(default="default a")
        child_key = Key(default="default b")
        child_key.setParent(parent_key)

        child_key.onValueChangeCall(stub)

        child_key.set("test 1")
        stub.assert_called_once_with("test 1")
        stub.reset_mock()

        child_key.clear()
        stub.assert_called_once_with("default a")
        stub.reset_mock()

        parent_key.set("test 2")
        stub.assert_called_once_with("test 2")
        stub.reset_mock()

        parent_key.clear()
        stub.assert_called_once_with("default a")

    def test_dump_fields(self) -> None:
        # An unattached Key should get dumped.

        key = Key("")
        assert list(key.dumpFields()) == []
        key.set("test")
        assert list(key.dumpFields()) == [("", "test")]

        class TestBunch(Bunch):
            str_key = Key(default="")
            serializable_key = Key(default=ExampleSerializable("empty"))
            _private = Key(default=0)

        bunch = TestBunch()

        # A public Key should get dumped.

        assert list(bunch.str_key.dumpFields()) == []
        bunch.str_key.set("test")
        assert list(bunch.str_key.dumpFields()) == [("", "test")]

        # A Key's value should be serialized in its dump.

        assert list(bunch.serializable_key.dumpFields()) == []
        bunch.serializable_key.set(ExampleSerializable("not empty"))
        assert list(bunch.serializable_key.dumpFields()) == [("", "not empty")]

        # A Key with a private label should not get dumped.

        assert list(bunch._private.dumpFields()) == []
        bunch._private.set(111)
        assert list(bunch._private.dumpFields()) == []

    def test_restore_field(self, mocker: MockerFixture) -> None:
        key: Key[int] = Key(0)
        callback = mocker.stub()
        key.onUpdateCall(callback)

        # The value should be set if the given label matches that of the Key. In
        # this case, "". In all other cases the value should not be set. As
        # well, restoring a field should not trigger a callback.

        assert not key.isSet()
        assert key.restoreFields([("", "1")])
        assert key.isSet()
        assert key.get() == 1
        callback.assert_not_called()

        key.clear()
        callback.reset_mock()
        assert not key.restoreFields([("invalid", "1")])
        assert not key.isSet()
        callback.assert_not_called()
        assert not key.restoreFields([(".invalid", "1")])
        assert not key.isSet()
        callback.assert_not_called()
        assert not key.restoreFields([("invalid.", "1")])
        assert not key.isSet()
        callback.assert_not_called()
        assert not key.restoreFields([(".", "1")])
        assert not key.isSet()
        callback.assert_not_called()

        # If multiple values are set, the first one sticks.

        key.clear()
        callback.reset_mock()
        assert key.restoreFields([("", "1"), ("", "2")])
        assert key.get() == 1
        callback.assert_not_called()

        # None is a valid value that clears the Key.

        assert key.isSet()
        assert key.restoreFields([("", None)])
        assert not key.isSet()
        callback.assert_not_called()

        # Restoring an empty list of fields clears the Key.

        key.set(123)
        assert key.isSet()
        callback.reset_mock()
        assert key.restoreFields([])
        assert not key.isSet()
        callback.assert_not_called()

        # Invalid values do not update the Key.

        key.clear()
        callback.reset_mock()
        assert not key.restoreFields([("", "?")])
        assert not key.isSet()
        callback.assert_not_called()
        assert not key.restoreFields([("", "")])
        assert not key.isSet()
        callback.assert_not_called()
        assert not key.restoreFields([("", None)])
        assert not key.isSet()
        callback.assert_not_called()

    def test_restore_field_serialization(self) -> None:
        key_str: Key[str] = Key(default="")
        assert key_str.restoreFields([("", "test")])
        assert key_str.get() == "test"

        key_int: Key[int] = Key(default=0)
        assert key_int.restoreFields([("", "12")])
        assert key_int.get() == 12

        key_float: Key[float] = Key(default=1.2)
        assert key_float.restoreFields([("", "3.4")])
        assert key_float.get() == 3.4

        key_bool: Key[bool] = Key(default=False)
        assert key_bool.restoreFields([("", "true")])
        assert key_bool.get()

        key_custom: Key[ExampleSerializable] = Key(default=ExampleSerializable(""))
        assert key_custom.restoreFields([("", "test")])
        assert key_custom.get().toStr() == "test"

    def test_invalid_restore_value_is_still_dumped(self) -> None:
        key = Key[int](default=0)

        assert not key.restoreFields([("", "12error")])
        assert key.get() == 0

        assert list(key.dumpFields()) == [("", "12error")]

        # Clearing or setting the Key also clears the bad value.

        key.clear()
        assert list(key.dumpFields()) == []

        assert not key.restoreFields([("", "12error")])
        key.set(0)
        assert list(key.dumpFields()) == [("", "0")]

    def test_persistence(self) -> None:
        # A Key does not keep a reference to its parent or children.

        key: Key[str] = Key(default="")
        level1: Key[str] = Key(default="")
        level1.setParent(key)
        level2: Key[str] = Key(default="")
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del key
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0

    def test_custom_serializer_passed_to_new_instances(self) -> None:
        class NeedsSerializer:
            pass

        class CustomSerializer:
            def toStr(self, value: NeedsSerializer) -> str:
                return "test"

            def fromStr(self, string: str) -> NeedsSerializer | None:
                return NeedsSerializer() if string == "custom" else None

        key = Key(default=NeedsSerializer(), serializer=CustomSerializer())
        other_key = key._newInstance()
        other_key.set(NeedsSerializer())

        assert list(other_key.dumpFields()) == [("", "test")]
        assert other_key.restoreFields([("", "custom")])
        assert not other_key.restoreFields([("", "invalid")])

    def test_complex_key_type_with_subclasses(self) -> None:
        class BaseClass:
            def toStr(self) -> str: ...

            @classmethod
            def fromStr(cls, string: str) -> "BaseClass": ...

        class Derived1(BaseClass):
            pass

        class Derived2(BaseClass):
            pass

        d1 = Derived1()
        key_broken: Key[BaseClass] = Key(default=d1)

        d2 = Derived2()
        assert not key_broken.set(d2)
        assert key_broken.get() is d1

        key_working: Key[BaseClass] = Key(default=d1, value_type=BaseClass)

        d2 = Derived2()
        assert key_working.set(d2)
        assert key_working.get() is d2

    def test_explicit_value_type_must_be_compatible_with_default(self) -> None:
        Key(default=0, value_type=int)

        with pytest.raises(TypeError):
            Key(default=0, value_type=str)

    def test_explicit_key_type_transmitted_to_new_instances(self) -> None:
        class BaseClass:
            def toStr(self) -> str: ...

            @classmethod
            def fromStr(cls, string: str) -> "BaseClass": ...

        class Derived1(BaseClass):
            pass

        class Derived2(BaseClass):
            pass

        key1: Key[BaseClass] = Key(default=Derived1(), value_type=BaseClass)

        key2 = key1._newInstance()
        d2 = Derived2()
        assert key2.set(d2)
        assert key2.get() is d2



from collections import defaultdict
from collections.abc import Callable, Iterable, MutableMapping
from typing import TypeVar

_T = TypeVar("_T")


def collate_by_prefix(
    fields: Iterable[tuple[str, _T]], splitter: Callable[[str], tuple[str, str]]
) -> MutableMapping[str, list[tuple[str, _T]]]:
    collated: dict[str, list[tuple[str, _T]]] = defaultdict(list)

    for path, value in fields:
        prefix, suffix = splitter(path)
        collated[prefix].append((suffix, value))

    return collated


def split_on(pivot: str) -> Callable[[str], tuple[str, str]]:
    def splitter(string: str) -> tuple[str, str]:
        before, after, *_ = [*string.split(pivot, maxsplit=1), ""]
        return before, after

    return splitter



import logging
import sys
from collections.abc import Callable, Iterable, MutableSet
from pathlib import Path
from typing import IO, Any

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import Self
else:
    from typing import Self

from sunset.autosaver import AutoSaver
from sunset.bunch import Bunch
from sunset.exporter import load_from_file, normalize, save_to_file
from sunset.lock import SettingsLock
from sunset.sets import NonHashableSet
from sunset.stringutils import collate_by_prefix, split_on

_MAIN = "main"

_FieldItemT = tuple[str, str | None]


class Settings(Bunch):
    """
    A collection of keys that can be saved to and loaded from text, and supports
    subsections.

    Under the hood, a Settings class is a dataclass, and can be used in the same
    manner, i.e. by defining attributes directly on the class itself.

    Settings instances support subsections: calling the :meth:`newSection()`
    method on an instance creates a subsection of that instance. This subsection
    holds the same keys, with independant values. If a key of the subsection
    does not have a value, its value will be looked up on its parent section
    instead. The hierarchy of sections can be arbitrarily deep.

    When saving a Settings instance, its subsections are saved with it under a
    distinct heading for each, provided they have a name. A section is given a
    name by passing it to the :meth:`newSection()` method, or by using the
    :meth:`setSectionName()` method on the new section after creation.

    The name of each section is used to construct the heading it is saved under.
    The top-level Settings instance is saved under the `[main]` heading by
    default.

    Anonymous (unnamed) sections do not get saved.

    Example:

    >>> from sunset import Key, Settings
    >>> class AnimalSettings(Settings):
    ...     hearts: Key[int] = Key(default=0)
    ...     legs: Key[int]   = Key(default=0)
    ...     wings: Key[int]  = Key(default=0)
    ...     fur: Key[bool]   = Key(default=False)
    >>> animals = AnimalSettings()
    >>> animals.hearts.set(1)
    True
    >>> mammals = animals.newSection(name="mammals")
    >>> mammals.fur.set(True)
    True
    >>> mammals.legs.set(4)
    True
    >>> humans = mammals.newSection(name="humans")
    >>> humans.legs.set(2)
    True
    >>> humans.fur.set(False)
    True
    >>> birds = animals.newSection(name="birds")
    >>> birds.legs.set(2)
    True
    >>> birds.wings.set(2)
    True
    >>> aliens = animals.newSection()  # No name given!
    >>> aliens.hearts.set(2)
    True
    >>> aliens.legs.set(7)
    True
    >>> print(mammals.hearts.get())
    1
    >>> print(mammals.legs.get())
    4
    >>> print(mammals.wings.get())
    0
    >>> print(mammals.fur.get())
    True
    >>> print(birds.hearts.get())
    1
    >>> print(birds.legs.get())
    2
    >>> print(birds.wings.get())
    2
    >>> print(birds.fur.get())
    False
    >>> print(humans.hearts.get())
    1
    >>> print(humans.legs.get())
    2
    >>> print(humans.wings.get())
    0
    >>> print(humans.fur.get())
    False
    >>> print(aliens.hearts.get())
    2
    >>> print(aliens.legs.get())
    7
    >>> print(aliens.wings.get())
    0
    >>> print(aliens.fur.get())
    False
    >>> import io
    >>> txt = io.StringIO()
    >>> animals.save(txt)
    >>> print(txt.getvalue(), end="")
    [main]
    hearts = 1
    [birds]
    legs = 2
    wings = 2
    [mammals]
    fur = true
    legs = 4
    [mammals/humans]
    fur = false
    legs = 2
    """

    MAIN: str = _MAIN

    _SECTION_SEPARATOR = "/"

    _section_name: str = ""
    _children_set: MutableSet[Bunch]
    _autosaver: AutoSaver | None = None
    _autosaver_class: type[AutoSaver]

    def __init__(self) -> None:
        super().__init__()

        # Note that this overrides the similarly named attribute from the parent
        # class. In the parent class, the set does not keep references to its
        # items; in this class, it does.

        self._children_set = NonHashableSet()
        self._autosaver_class = AutoSaver

    @SettingsLock.with_write_lock
    def newSection(self, name: str = "") -> Self:
        """
        Creates and returns a new instance of this class. Each key of the new
        instance will inherit from the key of the same name on the parent
        instance.

        When saving Settings with the :meth:`save()` method, each section's name
        is used to generate the heading under which that section is saved. If
        the new section is created without a name, it will be skipped when
        saving. A name can still be given to a section after creation with the
        :meth:`setSectionName()` method.

        If this Settings instance already has a section with the given name, the
        new section will be created with a unique name generated by appending a
        numbered suffix to that name.

        Args:
            name: The name that will be used to generate a heading for this
                section when saving it to text. The given name will be
                normalized to lowercase alphanumeric characters.

        Returns:
            An instance of the same type as self.
        """

        new = self._newInstance()
        new.setSectionName(name)

        # Note that this will trigger an update notification.

        new.setParent(self)

        return new

    @SettingsLock.with_write_lock
    def getOrCreateSection(self, name: str) -> Self:
        """
        Finds and returns the section of these Settings with the given name if
        it exists, and creates it if it doesn't.

        If the given name is empty, this is equivalent to calling
        :meth:`newSection()` instead.

        Args:
            name: The name that will be used to generate a heading for this
                section when saving it to text. The given name will be
                normalized to lowercase alphanumeric characters.

        Returns:
            An instance of the same type as self.
        """

        return (
            section
            if (section := self.getSection(name)) is not None
            else self.newSection(name=name)
        )

    @SettingsLock.with_read_lock
    def getSection(self, name: str) -> Self | None:
        """
        Finds and returns a section of this instance with the given name, if it
        exists, else None.

        Args:
            name: The name of the section to return.

        Returns:
            An instance of the same type as self, or None.
        """

        norm = normalize(name)
        if not norm:
            return None

        for section in self.sections():
            if norm == section.sectionName():
                return section

        return None

    @SettingsLock.with_read_lock
    def sections(self) -> Iterable[Self]:
        """
        Returns an iterable with the subsections of this Settings instance. Note that
        the subsections are only looked up one level deep, that is to say, no recursing
        into the section hierarchy occurs.

        Returns:
            An iterable of Settings instances of the same type as this one.
        """

        return sorted(self.children())

    @SettingsLock.with_write_lock
    def setSectionName(self, name: str) -> str:
        """
        Sets the unique name under which this Settings instance will be
        persisted by the :meth:`save()` method. The given name will be
        normalized to lowercase, without space or punctuation.

        This name is guaranteed to be unique. If the given name is already used
        by a section of the same Settings instance, then a numbered suffix is
        generated to make this one's name unique.

        If the given name is empty, these settings will be skipped when saving.

        The toplevel Settings instance is named "main" by default. It cannot be
        unnamed; it will revert to its default name instead.

        Args:
            name: The name that will be used to generate a heading for these
                settings when saving them to text.

        Returns:
            The given name, normalized, and made unique if needed.

        Example:

        >>> from sunset import Settings
        >>> class TestSettings(Settings):
        ...     pass
        >>> parent = TestSettings()
        >>> section1 = parent.newSection()
        >>> section2 = parent.newSection()
        >>> section3 = parent.newSection()
        >>> section1.setSectionName("  T ' e ? S / t")
        'test'
        >>> section2.setSectionName("TEST")
        'test_1'
        >>> section3.setSectionName("test")
        'test_2'
        >>> # This should not change this section's name.
        >>> section3.setSectionName("test")
        'test_2'
        """

        name = normalize(name)

        if (parent := self.parent()) is None:
            if name != self._section_name:
                self._section_name = name
                self._update_notifier.trigger(self)

        else:
            # Note that this triggers a notification if the unique name is
            # different from the current name.

            parent._setUniqueNameForSection(name, self)  # noqa: SLF001

        return self.sectionName()

    def _setUniqueNameForSection(self, name: str, section: Self) -> None:
        candidate = name = normalize(name)

        if candidate:
            other_names = {s.sectionName() for s in self.children() if s is not section}

            i = 0
            while candidate in other_names:
                i += 1
                candidate = f"{name}_{i}"

        if candidate != section._section_name:  # noqa: SLF001
            section._section_name = candidate  # noqa: SLF001
            section._update_notifier.trigger(section)  # noqa: SLF001

    @SettingsLock.with_read_lock
    def sectionName(self) -> str:
        """
        Returns the current name of this Settings instance. This name will be
        used to generate the heading under which this Settings instance will be
        persisted by the :meth:`save()` method.

        Returns:
            The name of this Settings instance.
        """
        if name := self._section_name:
            return name

        return name if self.parent() is not None else self.MAIN

    @SettingsLock.with_write_lock
    def setParent(self, parent: Self | None) -> None:  # type: ignore  # noqa: PGH003
        """
        Makes the given Settings instance the parent of this one. If None,
        remove this instance's parent, if any.

        All the Key, List and Bunch fields defined on this instance will be
        recursively reparented to the corresponding Key / List / Bunch field on
        the given parent.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Settings instance that will become this instance's
                parent, or None. The parent Settings instance must have the same
                type as this instance.
        """

        if parent is (previous_parent := self.parent()):
            return

        # Ensure that this section's name is unique in its parent.

        if parent is not None:
            # May trigger an update notification if the name is changed.

            parent._setUniqueNameForSection(self._section_name, self)  # noqa: SLF001
            self._update_notifier.add(parent._update_notifier.trigger)  # noqa: SLF001

        super().setParent(parent)

        if parent is not None:
            parent._update_notifier.trigger(parent)  # noqa: SLF001

        if previous_parent is not None:
            previous_parent._update_notifier.trigger(previous_parent)  # noqa: SLF001
            self._update_notifier.discard(previous_parent._update_notifier.trigger)  # noqa: SLF001

    # Not actually useless. This lets us override the docstring with
    # Settings-specific info.
    # pylint: disable-next=useless-parent-delegation
    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this Settings instance is updated.
        A Settings instance is considered updated when any of its fields is
        updated, or when its name is updated.

        The callback will be called with as its argument whichever field was
        just updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        super().onUpdateCall(callback)

    def skipOnSave(self) -> bool:
        return self.sectionName() == ""

    @SettingsLock.with_read_lock
    def dumpFields(self) -> Iterable[tuple[str, str | None]]:
        """
        Internal.
        """

        ret: list[tuple[str, str | None]] = []
        if not self.skipOnSave():
            # Ensure the section is dumped even if empty. Dumping an empty section is
            # valid.

            if not self.isSet():
                ret.append(("", None))
            else:
                ret.extend((path, item) for path, item in super().dumpFields())

            for section in self.sections():
                ret.extend(
                    (section.sectionName() + self._SECTION_SEPARATOR + path, item)
                    for path, item in section.dumpFields()
                )

        return ret

    @SettingsLock.with_write_lock
    def restoreFields(self, fields: Iterable[_FieldItemT]) -> bool:
        """
        Internal.
        """

        previous_subsections = {
            section.sectionName(): section for section in self.sections()
        }

        fields = list(fields)

        sep = self._SECTION_SEPARATOR
        bunch_fields = [(path, value) for path, value in fields if sep not in path]
        by_section = collate_by_prefix(
            [(path, value) for path, value in fields if sep in path],
            split_on(sep),
        )

        with self._update_notifier.inhibit():
            any_change = super().restoreFields(bunch_fields)

            for subsection_name, subsection in previous_subsections.items():
                if subsection_name not in by_section:
                    subsection.setParent(None)
                    subsection.clear()
                    any_change = True
                    continue

                any_change = (
                    subsection.restoreFields(by_section.pop(subsection_name))
                    or any_change
                )

            for subsection_name, subsection_fields in by_section.items():
                if not subsection_name:
                    continue
                subsection = self.getOrCreateSection(subsection_name)
                any_change = subsection.restoreFields(subsection_fields) or any_change

        return any_change

    def save(self, file: IO[str], *, blanklines: bool = False) -> None:
        """
        Writes the contents of this Settings instance and its subsections in
        text form to the given file object.

        Args:
            file: A text file object where to save this Settings instance.

            blanklines: Whether to add a blank line before section headings.
        """

        if self.skipOnSave():
            # This is an anonymous instance, actually. There is therefore
            # nothing to save.

            return

        save_to_file(
            file, self.dumpFields(), blanklines=blanklines, main=self.sectionName()
        )

    def load(self, file: IO[str]) -> None:
        """
        Loads settings from the given text file object.

        If the given file contains lines that don't make sense -- for instance,
        if the line is garbage, or refers to a key that does not exist in this
        Settings class, or it exists but with an incompatible type -- then the
        faulty line is skipped silently.

        If the text file object contains multiple headings, those headings will
        be used to create subsections with the corresponding names.

        Note that loading new settings resets the current settings.

        Args:
            file: A text file open in reading mode.
        """

        self.restoreFields(load_from_file(file, self.sectionName()))

    def autosave(
        self,
        path: str | Path,
        *,
        save_on_update: bool = True,
        save_delay: int = 0,
        raise_on_error: bool = False,
        logger: logging.Logger | None = None,
    ) -> AutoSaver:
        """
        Returns a context manager that loads these Settings from the given file
        path on instantiation and saves them on exit.

        By default, it will also automatically save the settings whenever they
        are updated from inside the application. Optionally, the updates can be
        batched over a given delay before being saved.

        See the documentation of :class:`~sunset.AutoSaver` for the details.

        Args:
            path: The full path to the file to load the settings from and save
                them to. If this file does not exist yet, it will be created
                when saving for the first time.

            save_on_update: Whether to save the settings when they
                are updated in any way. Default: True.

            save_delay: How long to wait, in seconds, before actually saving the
                settings when `save_on_update` is True and an update occurs.
                Setting this to a few seconds will batch updates for that long
                before triggering a save. If set to 0, the save is triggered
                immediately. Default: 0.

            raise_on_error: Whether OS errors occurring while loading and saving the
                settings should raise an exception. If False, errors will only be
                logged. Default: False.

            logger: A logger instance that will be used to log OS errors, if
                any, while loading or saving settings. If none is given, the
                default root logger will be used.

        Returns:
            An AutoSaver context manager.
        """

        # Keep a reference to the AutoSaver instance, so that its lifetime is
        # bound to that of this Settings instance.

        self._autosaver = self._autosaver_class(
            self,
            path,
            save_on_update=save_on_update,
            save_delay=save_delay,
            raise_on_error=raise_on_error,
            logger=logger,
        )
        return self._autosaver

    def __lt__(self, other: Self) -> bool:
        # Giving sections an order lets us easily sort them when dumping.

        return self.sectionName() < other.sectionName()



from collections.abc import Callable
import threading

import pytest

from sunset.lock import RWLock


def _start_threaded(target: Callable[[], None]) -> threading.Thread:
    task = threading.Thread(target=target, daemon=True)
    task.start()
    return task


class TestRWLock:
    def test_lock_is_reentrant(self) -> None:
        lock = RWLock()

        def try_lock() -> None:
            with lock.lock_reads():
                with lock.lock_reads():
                    pass

            with lock.lock_writes():
                with lock.lock_writes():
                    pass

            with lock.lock_reads():
                with lock.lock_writes():
                    pass

            with lock.lock_writes():
                with lock.lock_reads():
                    pass

        task = _start_threaded(target=try_lock)
        task.join(timeout=1.0)
        assert not task.is_alive()

    def test_concurrent_readers(self) -> None:
        reader_count = 100
        lock = RWLock()
        read_locks_all_held = threading.Barrier(reader_count + 1)

        def lock_and_wait() -> None:
            with lock.lock_reads():
                read_locks_all_held.wait()

        tasks = [_start_threaded(target=lock_and_wait) for _ in range(reader_count)]

        # This will block unless all threads are at the barrier, which proves that all
        # the readers are holding the read lock.
        assert read_locks_all_held.wait(timeout=1.0)

        [task.join(timeout=1.0) for task in tasks]
        assert all(not task.is_alive() for task in tasks)

    def test_readers_dont_block_readers(self) -> None:
        lock = RWLock()
        lock_acquired = threading.Event()
        testing_done = threading.Event()

        def hold_read_lock() -> None:
            with lock.lock_reads():
                lock_acquired.set()
                testing_done.wait()

        task = _start_threaded(target=hold_read_lock)

        lock_acquired.wait(timeout=1.0)
        assert lock._acquire_read_lock(blocking=False)
        lock._release_read_lock()

        testing_done.set()
        task.join(timeout=1.0)
        assert not task.is_alive()

    def test_writers_block_writers(self) -> None:
        lock = RWLock()
        lock_acquired = threading.Event()
        testing_done = threading.Event()

        def hold_write_lock() -> None:
            with lock.lock_writes():
                lock_acquired.set()
                testing_done.wait()

        task = _start_threaded(target=hold_write_lock)

        lock_acquired.wait(timeout=1.0)
        assert not lock._acquire_write_lock(blocking=False)

        testing_done.set()
        task.join(timeout=1.0)
        assert not task.is_alive()

    def test_readers_block_writers(self) -> None:
        lock = RWLock()
        lock_acquired = threading.Event()
        testing_done = threading.Event()

        def hold_read_lock() -> None:
            with lock.lock_reads():
                lock_acquired.set()
                testing_done.wait()

        task = _start_threaded(target=hold_read_lock)

        lock_acquired.wait(timeout=1.0)
        assert not lock._acquire_write_lock(blocking=False)

        testing_done.set()
        task.join(timeout=1.0)
        assert not task.is_alive()

    def test_writers_block_readers(self) -> None:
        lock = RWLock()
        lock_acquired = threading.Event()
        testing_done = threading.Event()

        def hold_write_lock() -> None:
            with lock.lock_writes():
                lock_acquired.set()
                testing_done.wait()

        task = _start_threaded(target=hold_write_lock)

        lock_acquired.wait(timeout=1.0)
        assert not lock._acquire_read_lock(blocking=False)

        testing_done.set()
        task.join(timeout=1.0)
        assert not task.is_alive()

    def test_readers_dont_starve_writers(self) -> None:
        reader_count = 100
        done = [False]

        lock = RWLock()
        read_locks_all_held = threading.Barrier(parties=reader_count + 1)
        write_lock_acquired = threading.Event()
        writer_can_quit = threading.Event()
        unpause = threading.Semaphore(value=0)

        def hold_read_lock_and_wait() -> None:
            while not done[0]:
                with lock.lock_reads():
                    read_locks_all_held.wait()
                    unpause.acquire()

        def try_write_lock() -> None:
            with lock.lock_writes():
                write_lock_acquired.set()
                writer_can_quit.wait()

        tasks = [_start_threaded(hold_read_lock_and_wait) for _ in range(reader_count)]

        # Ensure that all reader threads are in the locked section. Adding the final
        # waiter to the barrier will unblock the reader threads and they will
        # immediately block on the semaphore on the next line until we unpause it.
        read_locks_all_held.wait()

        # Start the writer. It should not be able to proceed because the reader lock is held.
        tasks.append(_start_threaded(try_write_lock))
        assert not write_lock_acquired.wait(timeout=0.01)
        assert not write_lock_acquired.is_set()

        # But if we let the reader threads proceed...
        unpause.release(reader_count)

        # Then the reader threads should not be able to re-acquire the read lock until
        # the writer waiting on the lock has acquired and then released it.
        assert write_lock_acquired.wait(1.0)

        # And while the writer holds the write lock, no reader should be able to
        # re-enter the critical section.
        assert read_locks_all_held.n_waiting == 0

        # Cleanup is a bit tricky because we need to unpause as many reader threads as
        # are paused. So we let the writer thread terminate and wait until all the
        # reader threads are in the critical section again...
        writer_can_quit.set()
        read_locks_all_held.wait()

        # Then signal that this loop should be the last...
        done[0] = True

        # And then let the reader threads proceed.
        unpause.release(reader_count)

        [task.join(timeout=1.0) for task in tasks]
        assert not any(task.is_alive() for task in tasks)

    def test_releasing_not_held_lock_raises_runtimeerror(self) -> None:
        lock = RWLock()
        lock._acquire_read_lock()
        lock._acquire_read_lock()
        lock._acquire_write_lock()
        lock._acquire_write_lock()

        lock._release_read_lock()
        lock._release_write_lock()
        lock._release_read_lock()
        lock._release_write_lock()

        with pytest.raises(RuntimeError):
            lock._release_read_lock()

        with pytest.raises(RuntimeError):
            lock._release_write_lock()



import enum
import functools
from typing import Generic, TypeVar, cast

_EnumT = TypeVar("_EnumT", bound=enum.Enum)


class EnumSerializer(Generic[_EnumT]):
    _type: type[_EnumT]

    def __init__(self, type_: type[_EnumT]) -> None:
        super().__init__()

        self._type = type_

    def _compute_and_set_name(self, value: _EnumT) -> None:
        if (
            not isinstance(value, enum.Flag)
            or getattr(value, "_name_", None) is not None
        ):
            return

        # WORKAROUND: In Python 3.10 and earlier, compound flags do not get
        # a name, which we need for serialization. So we have to compute one
        # ourselves.

        names: list[str] = []
        final_name: str | None = None
        current_value: int | None = None

        # Note how we sort the items by ascending value. This lets us keep
        # the name component order deterministic, as well as prioritize the
        # lowest-value members when compositing a compound flag.

        for name, member in sorted(
            self._members().items(), key=lambda item: item[1].value
        ):
            if member.value == value.value:
                # We found a member whose value is exactly this instance's
                # value. Let's use its name and end the process here.

                final_name = name
                break

            if member.value & value.value == member.value:
                # This member potentially contributes its value to the value
                # of this instance.

                if current_value is not None:
                    if current_value | member.value == current_value:
                        # This member does not add a contribution to the
                        # overall value of this instance that isn't already
                        # contributed by already accounted for members. Skip
                        # it.

                        continue

                    current_value |= member.value

                else:
                    current_value = member.value

                names.append(name)

        if names and not final_name:
            final_name = "|".join(names)

        if final_name:
            value._name_ = final_name

    def _members(self) -> dict[str, _EnumT]:
        # Note that we use the __members__ attribute and not dir() in order to get the
        # members in order of declaration.
        return {
            name: member
            for name, member in self._type.__members__.items()
            if isinstance(member, self._type)
        }

    def _lookupMember(self, name: str) -> _EnumT | None:
        candidates: list[_EnumT] = []

        for candidate_name, member in self._members().items():
            if candidate_name.strip() == name.strip():
                return member

            if candidate_name.strip().lower() == name.strip().lower():
                candidates.append(member)

        if len(candidates) == 1:
            return candidates[0]

        return None

    def fromStr(self, string: str) -> _EnumT | None:
        # value is either a single word, or, in the case of Flag enums, multiple
        # words separated by a pipe. We handle both cases together here.

        names = string.split("|") if "|" in string else [string]
        named_members = [
            member for name in names if (member := self._lookupMember(name)) is not None
        ]

        if not named_members:
            return None

        if issubclass(self._type, enum.Flag):
            # At this point, we know that all the members are instances of this
            # class, and this class is a subclass of enum.Flag. So we can safely
            # cast to make the type checker happy, apply the OR operator to the
            # members we found, and return the result.

            flag_members = cast(list[enum.Flag], named_members)
            ret = functools.reduce(lambda f1, f2: f1 | f2, flag_members)
            return cast(_EnumT, ret)

        # If this serializer is not used with a Flag, then there should be a
        # single member found on the class with the exact given name. If so,
        # return that, otherwise it's an error and we return None.

        return named_members[0] if len(named_members) == 1 else None

    def toStr(self, value: _EnumT) -> str:
        if getattr(value, "_name_", None) is None:
            # We need to call this here because the Enum internals in Python
            # 3.10 and earlier set the name to None *after* the instance is
            # created. So this call can't go in, say, __init__().

            self._compute_and_set_name(value)

        # Returning this instance's name property suffices due to the setup work
        # done in self._compute_and_set_name().

        return value.name


# LEGACY: In SunsetSettings 0.4, these custom Enum/Flag subclasses were made
# available. They are no longer necessary since we now can serialize Enums
# directly, but we keep providing them for backward compatibility.

SerializableEnum = enum.Enum
SerializableFlag = enum.Flag



from collections.abc import Callable
import threading
from typing import Any

from pytest import MonkeyPatch
from pytest_mock import MockerFixture

from sunset import PersistentTimer


class TestPersistentTimer:
    def test_timer_no_delay(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        mock_cls = mocker.MagicMock(threading.Timer, autospec=True)
        mock = mock_cls.return_value
        monkeypatch.setattr(threading, "Timer", mock_cls)

        def test() -> None: ...

        timer = PersistentTimer()

        assert timer._timer is None

        timer.start(0.0, test)

        mock_cls.assert_not_called()
        mock.start.assert_not_called()

        assert timer._timer is None

        timer.cancel()

        assert timer._timer is None

    def test_timer_with_delay(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        mock = mocker.MagicMock(threading.Timer, autospec=True)
        mock.return_value = mock
        monkeypatch.setattr(threading, "Timer", mock)

        def test() -> None: ...

        timer = PersistentTimer()

        assert timer._timer is None

        timer.start(1.0, test)

        mock.assert_called_once_with(1.0, timer._triggerFunction, [test])
        assert timer._timer is mock
        mock.start.assert_called_once()

        timer.cancel()

        assert timer._timer is None

    def test_timer_looping(
        self, mocker: MockerFixture, monkeypatch: MonkeyPatch
    ) -> None:
        timer_mock = mocker.MagicMock(threading.Timer, autospec=True)

        def mock_factory(
            interval: float, function: Callable[[], Any], args: list[Any]
        ) -> threading.Timer:
            timer_mock.interval = interval
            return timer_mock

        monkeypatch.setattr(threading, "Timer", mock_factory)

        called = mocker.Mock()

        persistent_timer = PersistentTimer(looping=True)

        assert persistent_timer._timer is None

        persistent_timer.start(1.0, called)

        assert persistent_timer._timer is timer_mock
        timer_mock.start.assert_called_once()
        timer_mock.start.reset_mock()

        called.assert_not_called()

        # Simulate the timer firing.
        persistent_timer._triggerFunction(called)

        # The callback should have been called.
        called.assert_called_once()

        # The timer should have restarted for a new loop.
        timer_mock.start.assert_called_once()

        persistent_timer.cancel()
        assert persistent_timer._timer is None


class MockTimer:
    # Not actually used in this file, but needed by other tests, so it makes sense to
    # have it defined somewhere semantically relevant.
    _clock: float
    _interval: float
    _function: Callable[[], Any] | None

    def __init__(self, *, looping: bool = False) -> None:
        self._function = None
        self._interval = 0
        self._clock = -1.0

    def cancel(self) -> None:
        self._clock = -1.0

    def start(self, interval: float, function: Callable[[], Any]) -> None:
        self._function = function
        if self._clock < 0:
            self._interval = interval
            self._clock = 0.0

    def advanceTime(self, time: float) -> None:
        if self._clock >= 0:
            self._clock += time
            if self._clock >= self._interval and self._function is not None:
                self._function()



import threading
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Any, Generic, ParamSpec

from sunset.sets import WeakCallableSet

_P = ParamSpec("_P")


class Notifier(Generic[_P]):
    """
    Records a set of callables of compatible types, and calls each of them with
    the given arguments when triggered.

    Recording a callable does not increase its reference count.
    """

    _callbacks: WeakCallableSet[Callable[_P, Any]]
    _inhibited: int = 0
    _lock: threading.Lock

    def __init__(self) -> None:
        self._callbacks = WeakCallableSet()
        self._lock = threading.Lock()

    def trigger(self, *args: _P.args, **kwargs: _P.kwargs) -> None:
        """
        Calls the recorded callables with the given arguments.

        Args:
            Anything compatible with the callable type recorded in this
            notifier.
        """

        if not self._inhibited:
            for callback in self._callbacks:
                callback(*args, **kwargs)

    def add(self, callback: Callable[_P, Any]) -> None:
        """
        Records a callable in this notifier.

        This does not increase the reference count of the callable.

        Args:
            callback: A callable to be recorded in this notifier.
        """

        self._callbacks.add(callback)

    def discard(self, callback: Callable[_P, Any]) -> None:
        """
        Forgets the given callable, if it was recorded in this notifier.

        If the callable was not previous recorded in this notifier, does
        nothing.

        Args:
            callback: The callable to be forgotten.
        """

        self._callbacks.discard(callback)

    @contextmanager
    def inhibit(self) -> Iterator[None]:
        """
        Returns a context manager that inhibits this notifier while active.

        An inhibited notifier no longer calls its recorded callable when triggered.

        Returns:
            A context manager. During its lifetime, triggering this notifier
            does nothing.
        """

        with self._lock:
            self._inhibited += 1
        try:
            yield None
        finally:
            with self._lock:
                self._inhibited -= 1
